from __future__ import print_function

import base64
import binascii
import json

import flask
import six
from six.moves import cPickle as pickle
from six.moves import urllib, zip

from smqtk.algorithms import (
    get_classifier_impls,
    get_descriptor_generator_impls,
    SupervisedClassifier
)
from smqtk.algorithms.classifier import (
    ClassifierCollection,
)
from smqtk.exceptions import MissingLabelError
from smqtk.iqr import IqrSession
from smqtk.representation import (
    ClassificationElementFactory,
    DescriptorElementFactory,
)
from smqtk.representation.data_element.memory_element import DataMemoryElement
import smqtk.utils.plugin
from smqtk.utils import prob_utils
from smqtk.utils.web import make_response_json
import smqtk.web


# Get expected JSON decode exception
# noinspection PyProtectedMember
if hasattr(flask.json._json, 'JSONDecodeError'):
    # noinspection PyProtectedMember
    JSON_DECODE_EXCEPTION = getattr(flask.json._json, 'JSONDecodeError')
else:
    # Exception thrown from ``json`` module.
    if six.PY2:
        JSON_DECODE_EXCEPTION = ValueError
    else:
        JSON_DECODE_EXCEPTION = json.JSONDecodeError


class SmqtkClassifierService (smqtk.web.SmqtkWebApp):
    """
    Headless web-app providing a RESTful API for classifying new data against
    a set of statically and dynamically loaded classifier models.

    The focus of this service is an endpoint where the user can send the
    base64-encoded data (with content type) they wish to be classified and get
    back the classification results of all loaded classifiers applied to the
    description of that data. Data for classification sent to this service is
    expected to be in

    Saved IQR session state bytes/files may be POST'ed to an endpoint with a
    descriptive label to add to the suite of classifiers that are run for
    user-provided data. The supervised classifier implementation that is
    trained from this state is part of the server configuration.

    Configuration Notes
    -------------------
    * The configured classifiers must all handle the descriptors output by the
      descriptor generator algorithm. IQR states loaded into the server must
      come from a service that also used the same descriptor generation
      algorithm. Otherwise the classification of new data will not make sense
      given the configured models as well as exceptions may occur due to
      descriptor dimensionality issues.

    * The classifier configuration provided for input IQR states should not
      have model persistence parameters specified since these classifiers will
      be ephemeral. If persistence parameters *are* specified, then subsequent
      IQR-state-based classifier models will bash each other causing
      erroneously labeled duplicate results.

    """

    CONFIG_ENABLE_CLASSIFIER_REMOVAL = "enable_classifier_removal"
    CONFIG_CLASSIFIER_COLLECTION = "classifier_collection"
    CONFIG_CLASSIFICATION_FACTORY = "classification_factory"
    CONFIG_DESCRIPTOR_GENERATOR = "descriptor_generator"
    CONFIG_DESCRIPTOR_FACTORY = "descriptor_factory"
    CONFIG_IMMUTABLE_LABELS = "immutable_labels"
    CONFIG_IQR_CLASSIFIER = "iqr_state_classifier_config"

    DEFAULT_IQR_STATE_CLASSIFIER_KEY = '__default__'

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        c = super(SmqtkClassifierService, cls).get_default_config()

        c[cls.CONFIG_ENABLE_CLASSIFIER_REMOVAL] = False

        # Static classifier configurations
        c[cls.CONFIG_CLASSIFIER_COLLECTION] = \
            ClassifierCollection.get_default_config()
        # Classification element factory for new classification results.
        c[cls.CONFIG_CLASSIFICATION_FACTORY] = \
            ClassificationElementFactory.get_default_config()
        # Descriptor generator for new content
        c[cls.CONFIG_DESCRIPTOR_GENERATOR] = smqtk.utils.plugin.make_config(
            get_descriptor_generator_impls()
        )
        # Descriptor factory for new content descriptors
        c[cls.CONFIG_DESCRIPTOR_FACTORY] = \
            DescriptorElementFactory.get_default_config()
        # from-IQR-state *supervised* classifier configuration
        c[cls.CONFIG_IQR_CLASSIFIER] = smqtk.utils.plugin.make_config(
            get_classifier_impls(
                sub_interface=SupervisedClassifier
            )
        )
        c[cls.CONFIG_IMMUTABLE_LABELS] = []

        return c

    def __init__(self, json_config):
        super(SmqtkClassifierService, self).__init__(json_config)

        self.enable_classifier_removal = \
            bool(json_config[self.CONFIG_ENABLE_CLASSIFIER_REMOVAL])

        self.immutable_labels = set(json_config[self.CONFIG_IMMUTABLE_LABELS])

        # Convert configuration into SMQTK plugin instances.
        #   - Static classifier configurations.
        #       - Skip the example config key
        #   - Classification element factory
        #   - Descriptor generator
        #   - Descriptor element factory
        #   - from-IQR-state classifier configuration
        #       - There must at least be the default key defined for when no
        #         specific classifier type is specified at state POST.

        # Classifier collection + factor
        self.classification_factory = \
            ClassificationElementFactory.from_config(
                json_config[self.CONFIG_CLASSIFICATION_FACTORY]
            )
        self.classifier_collection = ClassifierCollection.from_config(
            json_config[self.CONFIG_CLASSIFIER_COLLECTION]
        )

        # Descriptor generator + factory
        self.descriptor_factory = DescriptorElementFactory.from_config(
            json_config[self.CONFIG_DESCRIPTOR_FACTORY]
        )
        #: :type: smqtk.algorithms.DescriptorGenerator
        self.descriptor_gen = smqtk.utils.plugin.from_plugin_config(
            json_config[self.CONFIG_DESCRIPTOR_GENERATOR],
            smqtk.algorithms.get_descriptor_generator_impls()
        )

        # Classifier config for uploaded IQR states.
        self.iqr_state_classifier_config = \
            json_config[self.CONFIG_IQR_CLASSIFIER]

        self.add_routes()

    def add_routes(self):
        # REST API endpoint routes
        #
        # Example:
        # self.add_url_rule('/endpoint',
        #                   view_func=self.something,
        #                   methods=['GET'])
        #
        self.add_url_rule('/is_ready',
                          view_func=self.is_ready,
                          methods=['GET'])
        self.add_url_rule('/classifier_labels',
                          view_func=self.get_classifier_labels,
                          methods=['GET'])
        self.add_url_rule('/classifier_metadata',
                          view_func=self.get_classifier_metadata,
                          methods=['GET'])
        self.add_url_rule('/classify',
                          view_func=self.classify,
                          methods=['POST'])
        self.add_url_rule('/classifier',
                          view_func=self.get_classifier,
                          methods=['GET'])
        self.add_url_rule('/classifier',
                          view_func=self.add_classifier,
                          methods=['POST'])
        self.add_url_rule('/iqr_classifier',
                          view_func=self.add_iqr_state_classifier,
                          methods=['POST'])
        if self.enable_classifier_removal:
            self.add_url_rule('/classifier',
                              view_func=self.del_classifier,
                              methods=['DELETE'])

    # GET /is_ready
    # noinspection PyMethodMayBeStatic
    def is_ready(self):
        """
        Simple endpoint that just means this server is up and responding.
        """
        return make_response_json("Yes, I'm alive!")

    # GET /classifier_labels
    def get_classifier_labels(self):
        """
        Get the descriptive labels of the classifiers currently set to
        classify input data.

        Returns 200: {
            labels: list[str]
        }

        """
        all_labels = self.classifier_collection.labels()
        return make_response_json("Classifier labels.",
                                  labels=list(all_labels))

    # GET /classifier_metadata
    def get_classifier_metadata(self):
        """
        Get metadata associated with a specific classifier instance referred to
        by label.

        URL Arguments:
            label
                Reference label for a specific classifier to query.

        Returns code 200 on success and the JSON return object: {
            ...
            // Sequence of class labels that this classifier can classify
            // descriptors into.  This includes the negative label.
            class_labels=<list[str]>
        }

        """
        label = flask.request.values.get('label', default=None)
        if label is None or not label:
            return make_response_json("No label provided.", return_code=400)
        elif label not in self.classifier_collection.labels():
            return make_response_json("Label '%s' does not refer to a "
                                      "classifier currently registered."
                                      % label,
                                      return_code=404,
                                      label=label)
        class_labels = \
            self.classifier_collection.get_classifier(label).get_labels()
        return make_response_json("Success", return_code=200,
                                  class_labels=class_labels)

    # POST /classify
    def classify(self):
        """
        Given a file's bytes (standard base64-format) and content mimetype,
        describe and classify the content against all currently stored
        classifiers (optionally a list of requested classifiers), returning a
        map of classifier descriptive labels to their class-to-probability
        results.

        We expect the data to be transmitted in the body of the request in
        standard base64 encoding form ("bytes_b64" key). We look for the
        content type either as URL parameter or within the body
        ("content_type" key).

        Below is an example call to this endpoint via the ``requests`` python
        module, showing how base64 data is sent::

            import base64
            import requests
            data_bytes = "Load some content bytes here."
            requests.post('http://localhost:5000/classify',
                          data={'bytes_b64': base64.b64encode(data_bytes),
                                'content_type': 'text/plain'})

        With curl on the command line::

            $ curl -X POST localhost:5000/classify \
                -d "content_type=text/plain" \
                --data-urlencode "bytes_b64=$(base64 -w0 /path/to/file)"

            # If this fails, you may wish to encode the file separately and
            # use the file reference syntax instead:

            $ base64 -w0 /path/to/file > /path/to/file.b64
            $ curl -X POST localhost:5000/classify \
                -d "content_type=text/plain" \
                --data-urlencode bytes_64@/path/to/file.b64

        Optionally, the `label` parameter can be provided to limit the results
        of classification to a set of classifiers::

            $ curl -X POST localhost:5000/classify \
                -d "content_type=text/plain" \
                -d 'label=["some_label","other_label"]' \
                --data-urlencode "bytes_b64=$(base64 -w0 /path/to/file)"

            # If this fails, you may wish to encode the file separately and
            # use the file reference syntax instead:

            $ base64 -w0 /path/to/file > /path/to/file.b64
            $ curl -X POST localhost:5000/classify \
                -d "content_type=text/plain" \
                -d 'label=["some_label","other_label"]' \
                --data-urlencode bytes_64@/path/to/file.b64

        Data/Form arguments:
            bytes_b64
                Bytes in the standard base64 encoding to be described and
                classified.
            content_type
                The mimetype of the sent data.
            label
                (Optional) JSON-encoded label or list of labels
            adjustment
                (Optional) JSON-encoded dictionary of labels to floats. Higher
                values lower the gain on the class and therefore correspond to
                higher precision (and lower recall) for the class (and higher
                recall/lower precision for other classes). This translates git to
                calling ``smqtk.utils.prob_utils.adjust_proba``.

        Possible error codes:
            400
                No bytes provided, or provided labels are malformed
            404
                Label or labels provided do not match any registered
                classifier

        Returns: {
            ...
            result: {
                classifier-label: {
                    class-label: prob,
                    ...
                },
                ...
            }
        }

        """
        data_b64 = flask.request.values.get('bytes_b64', default=None)
        content_type = flask.request.values.get('content_type', default=None)
        label_str = flask.request.values.get('label', default=None)
        adjustment_str = flask.request.values.get('adjustment', default=None)

        labels = None
        if label_str is not None:
            try:
                labels = flask.json.loads(label_str)

                if isinstance(labels, six.string_types):
                    labels = [labels]
                elif isinstance(labels, list):
                    for el in labels:
                        if not isinstance(el, six.string_types):
                            return make_response_json(
                                "Label must be a list of strings or a"
                                " single string.", 400)
                else:
                    return make_response_json(
                        "Label must be a list of strings or a single"
                        " string.", 400)

            except JSON_DECODE_EXCEPTION:
                # Unquoted strings aren't valid JSON. That is, a plain string
                # needs to be passed as '"label"' rather than just 'label' or
                # "label". However, we can be a bit more generous and just
                # allow such a string, but we have to place *some* restriction
                # on it. We use `urllib.quote` for this since essentially it
                # just checks to make sure that the string is made up of one
                # of the following types of characters:
                #
                #   - letters
                #   - numbers
                #   - spaces, underscores, periods, and dashes
                #
                # Since the concept of a "letter" is fraught with encoding and
                # locality issues, we simply let urllib make this decision for
                # us.

                # If label_str matches the url-encoded version of itself, go
                # ahead and use it
                if urllib.parse.quote(label_str, safe='') == label_str:
                    labels = [label_str]
                else:
                    return make_response_json(
                        "Label(s) are not properly formatted JSON.", 400)

        # Collect optional result probability adjustment values
        #: :type: dict[collections.Hashable, float]
        adjustments = {}
        if adjustment_str is not None:
            try:
                #: :type: dict[collections.Hashable, float]
                adjustments = flask.json.loads(adjustment_str)

                for label, val in six.iteritems(adjustments):
                    if not isinstance(label, six.string_types):
                        return make_response_json(
                            "Adjustment label '%s' is not a string type."
                            % label,
                            400)
                    if not isinstance(val, (int, float)):
                        return make_response_json(
                            "Adjustment value %s for label '%s' is not an int "
                            "or float" % (val, label),
                            400)
            except JSON_DECODE_EXCEPTION:
                return make_response_json(
                    "Adjustment(s) are not properly formatted JSON.", 400)

        if data_b64 is None:
            return make_response_json("No base-64 bytes provided.", 400)
        elif content_type is None:
            return make_response_json("No content type provided.", 400)

        data_bytes = base64.b64decode(data_b64.encode('utf-8'))
        self._log.debug("Length of byte data: %d" % len(data_bytes))

        data_elem = DataMemoryElement(data_bytes, content_type, readonly=True)
        descr_elem = self.descriptor_gen.compute_descriptor(
            data_elem, self.descriptor_factory
        )
        self._log.debug("Descriptor shape: %s", descr_elem.vector().shape)

        try:
            clfr_map = self.classifier_collection.classify(
                descr_elem, labels=labels,
                factory=self.classification_factory)
        except MissingLabelError as ex:
            return make_response_json(
                "The following labels are not registered with any"
                " classifiers: '%s'"
                % "', '".join(ex.labels),
                404,
                missing_labels=list(ex.labels))

        # Transform classification result into JSON
        c_json = {}
        for classifier_label, c_elem in six.iteritems(clfr_map):
            prediction = c_elem.get_classification()
            if adjustments:
                proba_labels = list(prediction.keys())
                proba = [prediction[k] for k in proba_labels]
                # Use opposite of adjustments, because we already set the
                # convention of "higher: precision, lower: recall"
                adj = [-adjustments.get(label, 0.0) for label in proba_labels]
                adj_proba = prob_utils.adjust_proba(proba, adj)
                prediction = dict(zip(proba_labels, adj_proba[0]))
            c_json[classifier_label] = prediction

        return make_response_json('Finished classification.',
                                  result=c_json)

    # GET /classifier
    def get_classifier(self):
        """
        Download the classifier corresponding to the provided label, pickled
        and encoded in standard base64 encoding.

        Below is an example call to this endpoint via the ``requests`` python
        module::

            import base64
            import requests
            from six.moves import cPickle as pickle

            r = requests.get('http://localhost:5000/classifier',
                             data={'label': 'some_label'})
            data_bytes = base64.b64decode(r.content)
            classifier = pickle.loads(data_bytes)

        With curl on the command line::

            $ curl -X GET localhost:5000/classifier -d label=some_label | \
                base64 -d > /path/to/file.pkl

        Data args:
            label
                Label of the requested classifier

        Possible error codes:
            400
                No label provided
            404
                Label does not refer to a registered classifier

        Returns: The pickled and encoded classifier
        """
        label = flask.request.values.get('label', default=None)
        if label is None or not label:
            return make_response_json("No label provided.", 400)
        elif label not in self.classifier_collection.labels():
            return make_response_json("Label '%s' does not refer to a "
                                      "classifier currently registered."
                                      % label,
                                      404,
                                      label=label)

        clfr = self.classifier_collection.get_classifier(label)

        try:
            return base64.b64encode(pickle.dumps(clfr)), 200
        except pickle.PicklingError:
            return make_response_json("Classifier corresponding to label "
                                      "'%s' cannot be pickled." % label,
                                      500,
                                      label=label)

    # POST /iqr_classifier
    def add_iqr_state_classifier(self):
        """
        Train a classifier based on the user-provided IQR state file bytes in
        a base64 encoding, matched with a descriptive label of that
        classifier's topic.

        Since all IQR session classifiers end up only having two result
        classes (positive and negative), the topic of the classifier is
        encoded in the descriptive label the user applies to the classifier.

        Below is an example call to this endpoint via the ``requests`` python
        module, showing how base64 data is sent::

            import base64
            import requests
            data_bytes = "Load some content bytes here."
            requests.get('http://localhost:5000/iqr_classifier',
                         data={'bytes_b64': base64.b64encode(data_bytes),
                               'label': 'some_label'})

        With curl on the command line::

            $ curl -X POST localhost:5000/iqr_classifier \
                -d "label=some_label" \
                --data-urlencode "bytes_b64=$(base64 -w0 /path/to/file)"

            # If this fails, you may wish to encode the file separately and
            # use the file reference syntax instead:

            $ base64 -w0 /path/to/file > /path/to/file.b64
            $ curl -X POST localhost:5000/iqr_classifier -d label=some_label \
                --data-urlencode bytes_64@/path/to/file.b64

        To lock this classifier and guard it against deletion, add
        "lock_label=true"::

            $ curl -X POST localhost:5000/iqr_classifier \
                -d "label=some_label" \
                -d "lock_label=true" \
                --data-urlencode "bytes_b64=$(base64 -w0 /path/to/file)"

        Form arguments:
            iqr_state_b64
                base64 encoding of the bytes of the IQR session state save
                file.
            label
                Descriptive label to apply to this classifier. This should not
                conflict with existing classifier labels.
            lock_label
                If 'true', disallow deletion of this label. If 'false', allow
                deletion of this label. Only has an effect if deletion is
                enabled for this service. (Default: 'false')

        Returns 201.

        """
        data_b64 = flask.request.values.get('bytes_b64', default=None)
        label = flask.request.values.get('label', default=None)
        lock_clfr_str = flask.request.values.get('lock_label',
                                                 default='false')

        if data_b64 is None or len(data_b64) == 0:
            return make_response_json("No state base64 data provided.", 400)
        elif label is None or len(label) == 0:
            return make_response_json("No descriptive label provided.", 400)
        try:
            lock_clfr = bool(flask.json.loads(lock_clfr_str))
        except JSON_DECODE_EXCEPTION:
            return make_response_json("Invalid boolean value for"
                                      " 'lock_label'. Was given: '%s'"
                                      % lock_clfr_str,
                                      400)
        try:
            # Using urlsafe version because it handles both regular and urlsafe
            # alphabets.
            data_bytes = base64.urlsafe_b64decode(data_b64.encode('utf-8'))
        except (TypeError, binascii.Error) as ex:
            return make_response_json("Invalid base64 input: %s" % str(ex)), \
                   400

        # If the given label conflicts with one already in the collection,
        # fail.
        if label in self.classifier_collection.labels():
            return make_response_json(
                "Label already exists in classifier collection.", 400)

        # Create dummy IqrSession to extract pos/neg descriptors.
        iqrs = IqrSession()
        iqrs.set_state_bytes(data_bytes, self.descriptor_factory)
        pos = iqrs.positive_descriptors | iqrs.external_positive_descriptors
        neg = iqrs.negative_descriptors | iqrs.external_negative_descriptors
        del iqrs

        # Make a classifier instance from the stored config for IQR
        # session-based classifiers.
        #: :type: SupervisedClassifier
        classifier = smqtk.utils.plugin.from_plugin_config(
            self.iqr_state_classifier_config,
            get_classifier_impls(sub_interface=SupervisedClassifier)
        )
        classifier.train(class_examples={'positive': pos, 'negative': neg})

        try:
            self.classifier_collection.add_classifier(label, classifier)

            # If we're allowing deletions, get the lock flag from the form and
            # set it for this classifier
            if self.enable_classifier_removal and lock_clfr:
                self.immutable_labels.add(label)

        except ValueError as e:
            if e.args[0].find('JSON') > -1:
                return make_response_json("Tried to parse malformed JSON in "
                                          "form argument.", 400)
            return make_response_json("Duplicate label ('%s') added during "
                                      "classifier training of provided IQR "
                                      "session state." % label, 400,
                                      label=label)

        return make_response_json("Finished training IQR-session-based "
                                  "classifier for label '%s'." % label,
                                  201,
                                  label=label)

    # POST /classifier
    def add_classifier(self):
        """
        Upload a **trained** classifier pickled and encoded in standard base64
        encoding, matched with a descriptive label of that classifier's topic.

        Since all classifiers have only two result classes (positive and
        negative), the topic of the classifier is encoded in the descriptive
        label the user applies to the classifier.

        Below is an example call to this endpoint via the ``requests`` python
        module, showing how base64 data is sent::

            import base64
            import requests
            data_bytes = "Load some content bytes here."
            requests.post('http://localhost:5000/classifier',
                          data={'bytes_b64': base64.b64encode(data_bytes),
                                'label': 'some_label'})

        With curl on the command line::

            $ curl -X POST localhost:5000/classifier -d label=some_label \
                --data-urlencode "bytes_b64=$(base64 -w0 /path/to/file)"

            # If this fails, you may wish to encode the file separately and
            # use the file reference syntax instead:

            $ base64 -w0 /path/to/file.pkl > /path/to/file.pkl.b64
            $ curl -X POST localhost:5000/classifier -d label=some_label \
                --data-urlencode bytes_64@/path/to/file.pkl.b64

        To lock this classifier and guard it against deletion, add
        "lock_label=true"::

            $ curl -X POST localhost:5000/classifier \
                -d "label=some_label" \
                -d "lock_label=true" \
                --data-urlencode "bytes_b64=$(base64 -w0 /path/to/file.pkl)"

        Data/Form arguments:
            bytes_b64
                Bytes, in the standard base64 encoding, of the pickled
                classifier.
            label
                Descriptive label to apply to this classifier. This should not
                conflict with existing classifier labels.
            lock_label
                If 'true', disallow deletion of this label. If 'false', allow
                deletion of this label. Only has an effect if deletion is
                enabled for this service. (Default: 'false')

        Possible error codes:
            400
                May mean one of:
                    - No pickled classifier base64 data or label provided.
                    - Label provided is in conflict with an existing label in
                    the classifier collection.

        Returns code 201 on success and the message: {
            label: <str>
        }

        """
        clfr_b64 = flask.request.values.get('bytes_b64', default=None)
        label = flask.request.values.get('label', default=None)
        lock_clfr_str = flask.request.values.get('lock_label',
                                                 default='false')

        if clfr_b64 is None or len(clfr_b64) == 0:
            return make_response_json("No state base64 data provided.", 400)
        elif label is None or len(label) == 0:
            return make_response_json("No descriptive label provided.", 400)
        try:
            # This can throw a ValueError if lock_clfr is malformed JSON
            lock_clfr = bool(flask.json.loads(lock_clfr_str))
        except JSON_DECODE_EXCEPTION:
            return make_response_json("Invalid boolean value for"
                                      " 'lock_label'. Was given: '%s'"
                                      % lock_clfr_str,
                                      400)

        # If the given label conflicts with one already in the collection,
        # fail.
        if label in self.classifier_collection.labels():
            return make_response_json("Label '%s' already exists in"
                                      " classifier collection." % label,
                                      400,
                                      label=label)

        clfr = pickle.loads(base64.b64decode(clfr_b64.encode('utf-8')))

        try:
            self.classifier_collection.add_classifier(label, clfr)

            # If we're allowing deletions, get the lock flag from the form
            # and set it for this classifier
            if self.enable_classifier_removal and lock_clfr:
                self.immutable_labels.add(label)

        except ValueError:
            return make_response_json("Data added for label '%s' is not a"
                                      " Classifier." % label,
                                      400,
                                      label=label)

        return make_response_json("Uploaded classifier for label '%s'."
                                  % label,
                                  201,
                                  label=label)

    # DEL /classifier
    def del_classifier(self):
        """
        Remove a classifier by the given label.

        Form args:
            label
                Label of the classifier to remove.

        Possible error codes:
            400
                No classifier exists for the given label.

        Returns 200.

        """
        label = flask.request.values.get('label', default=None)
        if label is None or not label:
            return make_response_json("No label provided.", 400)
        elif label not in self.classifier_collection.labels():
            return make_response_json("Label '%s' does not refer to a"
                                      " classifier currently registered."
                                      % label,
                                      404,
                                      label=label)
        elif label in self.immutable_labels:
            return make_response_json("Label '%s' refers to a classifier"
                                      " that is immutable." % label,
                                      405,
                                      label=label)

        self.classifier_collection.remove_classifier(label)

        return make_response_json("Removed classifier with label '%s'."
                                  % label,
                                  removed_label=label)



"""
Script for asynchronously computing classifications for DescriptorElements
in a DescriptorIndex specified via a list of UUIDs. Results are output to a
CSV file in the format:

    uuid, label1_confidence, label2_confidence, ...

CSV columns labels are output to the given CSV header file path. Label
columns will be in the order as reported by the classifier implementations
``get_labels`` method.

Due to using an input file-list of UUIDs, we require that the UUIDs of
indexed descriptors be strings, or equality comparable to the UUIDs' string
representation.
"""

import csv
import logging
import os

from smqtk.algorithms import (
    get_classifier_impls
)
from smqtk.representation import (
    ClassificationElementFactory,
    get_classification_element_impls,
    get_descriptor_index_impls,
)
from smqtk.utils import (
    bin_utils,
    file_utils,
    parallel,
    plugin,
)


__author__ = "paul.tunison@kitware.com"


def default_config():
    return {
        "utility": {
            "classify_overwrite": False,
            "parallel": {
                "use_multiprocessing": False,
                "index_extraction_cores": None,
                "classification_cores": None,
            }
        },
        "plugins": {
            "classifier": plugin.make_config(get_classifier_impls()),
            "classification_factory": plugin.make_config(
                get_classification_element_impls()
            ),
            "descriptor_index": plugin.make_config(
                get_descriptor_index_impls()
            ),
        }
    }


def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)

    g_io = parser.add_argument_group("Input Output Files")
    g_io.add_argument('--uuids-list', metavar='PATH',
                      help='Path to the input file listing UUIDs to process.')
    g_io.add_argument('--csv-header', metavar='PATH',
                      help='Path to the file to output column header labels.')
    g_io.add_argument('--csv-data', metavar='PATH',
                      help='Path to the file to output the CSV data to.')
    return parser


def main():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    # - parallel_map UUIDs to load from the configured index
    # - classify iterated descriptors

    uuids_list_filepath = args.uuids_list
    output_csv_filepath = args.csv_data
    output_csv_header_filepath = args.csv_header
    classify_overwrite = config['utility']['classify_overwrite']

    p_use_multiprocessing = \
        config['utility']['parallel']['use_multiprocessing']
    p_index_extraction_cores = \
        config['utility']['parallel']['index_extraction_cores']
    p_classification_cores = \
        config['utility']['parallel']['classification_cores']

    if not uuids_list_filepath:
        raise ValueError("No uuids_list_filepath specified.")
    elif not os.path.isfile(uuids_list_filepath):
        raise ValueError("Given uuids_list_filepath did not point to a file.")
    if output_csv_header_filepath is None:
        raise ValueError("Need a path to save CSV header labels")
    if output_csv_filepath is None:
        raise ValueError("Need a path to save CSV data.")

    #
    # Initialize configured plugins
    #

    log.info("Initializing descriptor index")
    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(
        config['plugins']['descriptor_index'],
        get_descriptor_index_impls()
    )

    log.info("Initializing classification factory")
    c_factory = ClassificationElementFactory.from_config(
        config['plugins']['classification_factory']
    )

    log.info("Initializing classifier")
    #: :type: smqtk.algorithms.Classifier
    classifier = plugin.from_plugin_config(
        config['plugins']['classifier'], get_classifier_impls()
    )

    #
    # Setup/Process
    #
    def iter_uuids():
        with open(uuids_list_filepath) as f:
            for l in f:
                yield l.strip()

    def descr_for_uuid(uuid):
        """
        :type uuid: collections.Hashable
        :rtype: smqtk.representation.DescriptorElement
        """
        return descriptor_index.get_descriptor(uuid)

    def classify_descr(d):
        """
        :type d: smqtk.representation.DescriptorElement
        :rtype: smqtk.representation.ClassificationElement
        """
        return classifier.classify(d, c_factory, classify_overwrite)

    log.info("Initializing uuid-to-descriptor parallel map")
    #: :type: collections.Iterable[smqtk.representation.DescriptorElement]
    element_iter = parallel.parallel_map(
        descr_for_uuid, iter_uuids(),
        use_multiprocessing=p_use_multiprocessing,
        cores=p_index_extraction_cores,
        name="descr_for_uuid",
    )

    log.info("Initializing descriptor-to-classification parallel map")
    #: :type: collections.Iterable[smqtk.representation.ClassificationElement]
    classification_iter = parallel.parallel_map(
        classify_descr, element_iter,
        use_multiprocessing=p_use_multiprocessing,
        cores=p_classification_cores,
        name='classify_descr',
    )

    #
    # Write/Output files
    #

    c_labels = classifier.get_labels()

    def make_row(e):
        """
        :type e: smqtk.representation.ClassificationElement
        """
        c_m = e.get_classification()
        return [e.uuid] + [c_m[l] for l in c_labels]

    # column labels file
    log.info("Writing CSV column header file: %s", output_csv_header_filepath)
    file_utils.safe_create_dir(os.path.dirname(output_csv_header_filepath))
    with open(output_csv_header_filepath, 'wb') as f_csv:
        w = csv.writer(f_csv)
        w.writerow(['uuid'] + [str(cl) for cl in c_labels])

    # CSV file
    log.info("Writing CSV data file: %s", output_csv_filepath)
    file_utils.safe_create_dir(os.path.dirname(output_csv_filepath))
    r_state = [0] * 7
    with open(output_csv_filepath, 'wb') as f_csv:
        w = csv.writer(f_csv)
        for c in classification_iter:
            w.writerow(make_row(c))
            bin_utils.report_progress(log.info, r_state, 1.0)

    # Final report
    r_state[1] -= 1
    bin_utils.report_progress(log.info, r_state, 0)

    log.info("Done")


if __name__ == '__main__':
    main()



"""
Based on an input, trained classifier configuration, classify a number of media
files, whose descriptor is computed by the configured descriptor generator.
Input files that classify as the given label are then output to standard out.
Thus, this script acts like a filter.
"""
from __future__ import print_function

import glob
import json
import logging
import os

import six

from smqtk.algorithms import get_classifier_impls
from smqtk.algorithms import get_descriptor_generator_impls

from smqtk.representation import ClassificationElementFactory
from smqtk.representation import DescriptorElementFactory
from smqtk.representation.data_element.file_element import DataFileElement

from smqtk.utils import plugin
from smqtk.utils.bin_utils import (
    initialize_logging,
    output_config,
    basic_cli_parser,
)


def get_cli_parser():
    parser = basic_cli_parser(__doc__)

    g_classifier = parser.add_argument_group("Classification")
    g_classifier.add_argument('--overwrite',
                              action='store_true', default=False,
                              help='When generating a configuration file, '
                                   'overwrite an existing file.')
    g_classifier.add_argument('-l', '--label',
                              type=str, default=None,
                              help='The class to filter by. This is based on '
                                   'the classifier configuration/model used. '
                                   'If this is not provided, we will list the '
                                   'available labels in the provided '
                                   'classifier configuration.')

    # Positional
    parser.add_argument("file_globs",
                        nargs='*',
                        metavar='GLOB',
                        help='Series of shell globs specifying the files to '
                             'classify.')

    return parser


def get_default_config():
    return {
        "descriptor_factory":
            DescriptorElementFactory.get_default_config(),
        "descriptor_generator":
            plugin.make_config(get_descriptor_generator_impls()),
        "classification_factory":
            ClassificationElementFactory.get_default_config(),
        "classifier":
            plugin.make_config(get_classifier_impls()),
    }


def main():
    log = logging.getLogger(__name__)
    parser = get_cli_parser()
    args = parser.parse_args()

    config_path = args.config
    generate_config = args.generate_config
    config_overwrite = args.overwrite
    is_debug = args.verbose

    label = args.label
    file_globs = args.file_globs

    initialize_logging(logging.getLogger('__main__'),
                       is_debug and logging.DEBUG or logging.INFO)
    initialize_logging(logging.getLogger('smqtk'),
                       is_debug and logging.DEBUG or logging.INFO)
    log.debug("Showing debug messages.")

    config = get_default_config()
    config_loaded = False
    if config_path and os.path.isfile(config_path):
        with open(config_path) as f:
            log.info("Loading configuration: %s", config_path)
            config.update(
                json.load(f)
            )
        config_loaded = True
    output_config(generate_config, config, log, config_overwrite, 100)

    if not config_loaded:
        log.error("No configuration provided")
        exit(101)

    classify_files(config, label, file_globs)


def classify_files(config, label, file_globs):
    log = logging.getLogger(__name__)

    #: :type: smqtk.algorithms.Classifier
    classifier = \
        plugin.from_plugin_config(config['classifier'],
                                  get_classifier_impls())

    def log_avaialable_labels():
        log.info("Available classifier labels:")
        for l in classifier.get_labels():
            log.info("- %s", l)

    if label is None:
        log_avaialable_labels()
        return
    elif label not in classifier.get_labels():
        log.error("Invalid classification label provided to compute and filter "
                  "on: '%s'", label)
        log_avaialable_labels()
        return

    log.info("Collecting files from globs")
    #: :type: list[DataFileElement]
    data_elements = []
    uuid2filepath = {}
    for g in file_globs:
        if os.path.isfile(g):
            d = DataFileElement(g)
            data_elements.append(d)
            uuid2filepath[d.uuid()] = g
        else:
            log.debug("expanding glob: %s", g)
            for fp in glob.iglob(g):
                d = DataFileElement(fp)
                data_elements.append(d)
                uuid2filepath[d.uuid()] = fp
    if not data_elements:
        raise RuntimeError("No files provided for classification.")

    log.info("Computing descriptors")
    descriptor_factory = \
        DescriptorElementFactory.from_config(config['descriptor_factory'])
    #: :type: smqtk.algorithms.DescriptorGenerator
    descriptor_generator = \
        plugin.from_plugin_config(config['descriptor_generator'],
                                  get_descriptor_generator_impls())
    descr_map = descriptor_generator\
        .compute_descriptor_async(data_elements, descriptor_factory)

    log.info("Classifying descriptors")
    classification_factory = ClassificationElementFactory \
        .from_config(config['classification_factory'])
    classification_map = classifier\
        .classify_async(list(descr_map.values()), classification_factory)

    log.info("Printing input file paths that classified as the given label.")
    # map of UUID to filepath:
    uuid2c = dict((c.uuid, c) for c in six.itervalues(classification_map))
    for data in data_elements:
        d_uuid = data.uuid()
        log.debug("'{}' classification map: {}".format(
            uuid2filepath[d_uuid], uuid2c[d_uuid].get_classification()
        ))
        if uuid2c[d_uuid].max_label() == label:
            print(uuid2filepath[d_uuid])


if __name__ == '__main__':
    main()



import numpy as np

from smqtk.algorithms.descriptor_generator import DescriptorGenerator


class DummyDescriptorGenerator (DescriptorGenerator):
    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this
        class's ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        # No constructor, no config
        return {}

    @classmethod
    def is_usable(cls):
        """
        Check whether this class is available for use.

        Since certain plugin implementations may require additional
        dependencies that may not yet be available on the system, this method
        should check for those dependencies and return a boolean saying if the
        implementation is usable.

        NOTES:
            - This should be a class method
            - When an implementation is deemed not usable, this should emit a
                warning detailing why the implementation is not available for
                use.

        :return: Boolean determination of whether this implementation is
                 usable.
        :rtype: bool

        """
        return True

    def valid_content_types(self):
        """
        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """
        return ['text/plain']

    def _compute_descriptor(self, data):
        """
        Internal method that defines the generation of the descriptor vector
        for a given data element. This returns a numpy array.

        This method is only called if the data element has been verified to be
        of a valid content type for this descriptor implementation.

        :raises RuntimeError: Feature extraction failure of some kind.

        :param data: Some kind of input data for the feature descriptor.
        :type data: smqtk.representation.DataElement

        :return: Feature vector.
        :rtype: numpy.core.multiarray.ndarray

        """
        return np.zeros((5,), np.float64)



from six.moves import cPickle

from smqtk.exceptions import NoClassificationError
from smqtk.representation import ClassificationElement


# Try to import required modules
try:
    import psycopg2
except ImportError:
    psycopg2 = None


__author__ = "paul.tunison@kitware.com"


class PostgresClassificationElement (ClassificationElement):
    """
    PostgreSQL database backed classification element.

    Requires a table of at least 3 fields (column names configurable):

        - type-name :: text
        - uuid :: text
        - classification-binary :: bytea

    """

    UPSERT_TABLE_TMPL = ' '.join("""
        CREATE TABLE IF NOT EXISTS {table_name:s} (
          {type_col:s} TEXT NOT NULL
          {uuid_col:s} TEXT NOT NULL,
          {classification_col:s} BYTEA NOT NULL,
          PRIMARY KEY ({type_col:s}, {uuid_col:s})
        );
    """.split())

    # Known psql version compatibility: 9.4
    SELECT_TMPL = ' '.join("""
        SELECT {classification_col:s}
          FROM {table_name:s}
          WHERE {type_col:s} = %(type_val)s
            AND {uuid_col:s} = %(uuid_val)s
        ;
    """.split())

    # Known psql version compatibility: 9.4
    UPSERT_TMPL = ' '.join("""
        WITH upsert AS (
          UPDATE {table_name:s}
            SET {classification_col:s} = %(classification_val)s
            WHERE {type_col:s} = %(type_val)s
              AND {uuid_col:s} = %(uuid_val)s
            RETURNING *
          )
        INSERT INTO {table_name:s}
          ({type_col:s}, {uuid_col:s}, {classification_col:s})
          SELECT %(type_val)s, %(uuid_val)s, %(classification_val)s
            WHERE NOT EXISTS (SELECT * FROM upsert);
    """.split())

    @classmethod
    def is_usable(cls):
        if psycopg2 is None:
            cls.get_logger().warning("Not usable. Requires psycopg2 module")
            return False
        return True

    def __init__(self, type_name, uuid,
                 table_name='classifications',
                 type_col='type_name', uuid_col='uid',
                 classification_col='classification',
                 db_name='postgres', db_host=None, db_port=None, db_user=None,
                 db_pass=None, pickle_protocol=-1, create_table=True):
        """
        Initialize new PostgresClassificationElement attached to some database
        credentials.

        We require that storage tables treat uuid AND type string columns as
        primary keys. The type and uuid columns should be of the ``text`` type.
        The binary column should be of the ``bytea`` type.

        Default argument values assume a local PostgreSQL database with a table
        created via the
        ``etc/smqtk/postgres/classification_element/example_table_init.sql``
        file (relative to the SMQTK source tree or install root).

        NOTES:
            - Not all uuid types used here are necessarily of the ``uuid.UUID``
              type, thus the recommendation to use a ``text`` type for the
              column. For certain specific use cases they may be proper
              ``uuid.UUID`` instances or strings, but this cannot be generally
              assumed.

        :param type_name: Name of the type of classifier this classification was
            generated by.
        :type type_name: str

        :param uuid: Unique ID reference of the classification
        :type uuid: collections.Hashable

        :param table_name: String label of the database table to use.
        :type table_name: str

        :param uuid_col: The column label for classification UUID storage
        :type uuid_col: str

        :param type_col: The column label for classification type name storage.
        :type type_col: str

        :param classification_col: The column label for classification binary
            storage.
        :type classification_col: str

        :param db_host: Host address of the Postgres server. If None, we
            assume the server is on the local machine and use the UNIX socket.
            This might be a required field on Windows machines (not tested yet).
        :type db_host: str | None

        :param db_port: Port the Postgres server is exposed on. If None, we
            assume the default port (5423).
        :type db_port: int | None

        :param db_name: The name of the database to connect to.
        :type db_name: str

        :param db_user: Postgres user to connect as. If None, postgres
            defaults to using the current accessing user account name on the
            operating system.
        :type db_user: str | None

        :param db_pass: Password for the user we're connecting as. This may be
            None if no password is to be used.
        :type db_pass: str | None

        :param pickle_protocol: Pickling protocol to use. We will use -1 by
            default (latest version, probably binary).
        :type pickle_protocol: int

        :param create_table: If this instance should try to create the storing
            table before actions are performed against it. If the configured
            user does not have sufficient permissions to create the table and it
            does not currently exist, an exception will be raised.
        :type create_table: bool

        """
        super(PostgresClassificationElement, self).__init__(type_name, uuid)

        self.table_name = table_name
        self.type_col = type_col
        self.uuid_col = uuid_col
        self.classification_col = classification_col

        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port
        self.db_user = db_user
        self.db_pass = db_pass

        self.pickle_protocol = pickle_protocol
        self.create_table = create_table

    def get_config(self):
        return {
            "table_name": self.table_name,
            "type_col": self.type_col,
            "uuid_col": self.uuid_col,
            "classification_col": self.classification_col,

            "db_name": self.db_name,
            "db_host": self.db_host,
            "db_port": self.db_port,
            "db_user": self.db_user,
            "db_pass": self.db_pass,

            "pickle_protocol": self.pickle_protocol,
            "create_table": self.create_table,
        }

    def _get_psql_connection(self):
        """
        :return: A new connection to the configured database
        :rtype: psycopg2._psycopg.connection
        """
        return psycopg2.connect(
            database=self.db_name,
            user=self.db_user,
            password=self.db_pass,
            host=self.db_host,
            port=self.db_port,
        )

    def _ensure_table(self, cursor):
        """
        Execute on psql connector cursor the table create-of-not-exists query.

        :param cursor: Connection active cursor.

        """
        if self.create_table:
            q_table_upsert = self.UPSERT_TABLE_TMPL.format(**dict(
                table_name=self.table_name,
                type_col=self.type_col,
                uuid_col=self.uuid_col,
                classification_col=self.classification_col,
            ))
            cursor.execute(q_table_upsert)

    def has_classifications(self):
        """
        :return: If this element has classification information set.
        :rtype: bool
        """
        try:
            return bool(self.get_classification())
        except NoClassificationError:
            return False

    def get_classification(self):
        """
        Get classification result map, returning a label-to-confidence dict.

        We do no place any guarantees on label value types as they may be
        represented in various forms (integers, strings, etc.).

        Confidence values are in the [0,1] range.

        :raises NoClassificationError: No classification labels/confidences yet
            set.

        :return: Label-to-confidence dictionary.
        :rtype: dict[collections.Hashable, float]

        """
        q_select = self.SELECT_TMPL.format(**dict(
            table_name=self.table_name,
            type_col=self.type_col,
            uuid_col=self.uuid_col,
            classification_col=self.classification_col,
        ))
        q_select_values = {
            "type_val": self.type_name,
            "uuid_val": str(self.uuid)
        }

        conn = self._get_psql_connection()
        cur = conn.cursor()
        try:
            self._ensure_table(cur)
            cur.execute(q_select, q_select_values)
            r = cur.fetchone()
            # For server cleaning (e.g. pgbouncer)
            conn.commit()

            if not r:
                raise NoClassificationError("No PSQL backed classification for "
                                            "label='%s' uuid='%s'"
                                            % (self.type_name, str(self.uuid)))
            else:
                b = r[0]
                c = cPickle.loads(str(b))
                return c
        except Exception:
            conn.rollback()
            raise
        finally:
            cur.close()
            conn.close()

    def set_classification(self, m=None, **kwds):
        """
        Set the whole classification map for this element. This will strictly
        overwrite the entire label-confidence mapping (vs. updating it)

        Label/confidence values may either be provided via keyword arguments or
        by providing a dictionary mapping labels to confidence values.

        :param m: New labels-to-confidence mapping to set.
        :type m: dict[collections.Hashable, float]

        :raises ValueError: The given label-confidence map was empty.

        """
        m = super(PostgresClassificationElement, self)\
            .set_classification(m, **kwds)

        q_upsert = self.UPSERT_TMPL.strip().format(**{
            "table_name": self.table_name,
            "classification_col": self.classification_col,
            "type_col": self.type_col,
            "uuid_col": self.uuid_col,
        })
        q_upsert_values = {
            "classification_val":
                psycopg2.Binary(cPickle.dumps(m, self.pickle_protocol)),
            "type_val": self.type_name,
            "uuid_val": str(self.uuid),
        }

        conn = self._get_psql_connection()
        cur = conn.cursor()
        try:
            self._ensure_table(cur)
            cur.execute(q_upsert, q_upsert_values)
            cur.close()
            conn.commit()
        except Exception:
            conn.rollback()
            raise
        finally:
            cur.close()
            conn.close()



import logging
from multiprocessing import RLock
import uuid

from smqtk.utils import SmqtkObject

try:
    import psycopg2
    import psycopg2.pool
except ImportError as ex:
    logging.getLogger(__name__)\
           .warning("Failed to import psycopg2: %s", str(ex))
    psycopg2 = None


GLOBAL_PSQL_TABLE_CREATE_RLOCK = RLock()
GLOBAL_CONNECTION_POOL_LOCK = RLock()


# This dictionary contains a set of global connection pools. Each key is a
# tuple of a database name, host, user, and port. The value is the connection
# pool for that tuple. If we didn't do it this way, each PsqlConnectionHelper
# would have its own connection pool, even if it had the same credentials,
# which would defeat the purpose of pooling.
_connection_pools = dict()


def get_connection_pool(db_name, db_host, db_port, db_user, db_pass):
    """
    Get the connection pool instance for a specific address specification.

    :param db_name: The name of the database to connect to.
    :type db_name: str

    :param db_host: Host address of the Postgres server. If None, we
        assume the server is on the local machine and use the UNIX socket.
        This might be a required field on Windows machines (not tested yet).
    :type db_host: str | None

    :param db_port: Port the Postgres server is exposed on. If None, we
        assume the default port (5423).
    :type db_port: int | None

    :param db_user: Postgres user to connect as. If None, postgres
        defaults to using the current accessing user account name on the
        operating system.
    :type db_user: str | None

    :param db_pass: Password for the user we're connecting as. This may be
        None if no password is to be used.
    :type db_pass: str | None

    :return: New or existing connection pool instance for the given address
        specification.
    :rtype: psycopg2.pool.ThreadedConnectionPool
    """
    key_tuple = (
        db_name,
        db_host,
        db_port,
        db_user,
    )
    with GLOBAL_CONNECTION_POOL_LOCK:
        try:
            cp = _connection_pools[key_tuple]
        except KeyError:
            _connection_pools[key_tuple] = cp = \
                psycopg2.pool.ThreadedConnectionPool(
                    # FIXME: The min and max connections have been
                    # hard-coded to sensible values, but we should find a
                    # way to make them configurable.
                    0, 100,
                    database=db_name,
                    user=db_user,
                    password=db_pass,
                    host=db_host,
                    port=db_port,
                )
    return cp


def norm_psql_cmd_string(s):
    """
    Simple function to reduce down a multi-line string written for readability
    to a single line.

    :param s: Single or multi-line string with inconsistant white-space.
    :type s: str

    :return: Single-line string.
    :rtype: str

    """
    return ' '.join(s.split())


class PsqlConnectionHelper (SmqtkObject):
    """
    Helper class for things that interact with a PostgreSQL database.
    """

    def __init__(self, db_name='postgres', db_host=None, db_port=None,
                 db_user=None, db_pass=None,
                 itersize=1000,
                 table_upsert_lock=GLOBAL_PSQL_TABLE_CREATE_RLOCK):
        """
        Create a new helper instance for a set of database connection
        parameters.

        :param db_name: The name of the database to connect to.
        :type db_name: str

        :param db_host: Host address of the Postgres server. If None, we
            assume the server is on the local machine and use the UNIX socket.
            This might be a required field on Windows machines (not tested yet).
        :type db_host: str | None

        :param db_port: Port the Postgres server is exposed on. If None, we
            assume the default port (5423).
        :type db_port: int | None

        :param db_user: Postgres user to connect as. If None, postgres
            defaults to using the current accessing user account name on the
            operating system.
        :type db_user: str | None

        :param db_pass: Password for the user we're connecting as. This may be
            None if no password is to be used.
        :type db_pass: str | None

        :param itersize: Number of records fetched per network round trip when
            iterating over a named cursor. This parameter only does anything if
            a named cursor is used.
        :type itersize: int

        :param table_upsert_lock: Lock to guard table upsertion.
        :type table_upsert_lock: multiprocessing.synchronize.RLock
        """
        # Stop construction if psycopg2 is not defined.
        if psycopg2 is None:
            raise RuntimeError("Cannot construct a %s if the psycopg2 module "
                               "is not importable" % self.__class__.__name__)
        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port
        self.db_user = db_user
        self.db_pass = db_pass

        self.itersize = itersize

        self.table_upsert_lock = table_upsert_lock
        self.table_upsert_sql = None

        self.connection_pool = get_connection_pool(db_name, db_host, db_port,
                                                   db_user, db_pass)

    def get_psql_connection(self):
        """
        :return: A new connection to the configured database
        :rtype: psycopg2._psycopg.connection
        """
        return self.connection_pool.getconn()

    @staticmethod
    def get_unique_cursor_name():
        """
        Return a string to use as a cursor name.

        Named cursors are used to allow for server-side iteration for SELECT
        and VALUES commands. This lessens the load on the machine making the
        query.

        :return: New cursor name string with a unique, random UUID embedded.
        :rtype: str
        """
        ruuid = str(uuid.uuid4()).replace('-', '')
        return "smqtk_postgres_cursor_%s" % ruuid

    def set_table_upsert_sql(self, s):
        """
        SQL optional statement to upsert a table in the database before
        executing statements. If this is not set, table upsertion will not
        occur.

        ``None`` may be to disable upsertion.

        :param s: String SQL statement or ``None``.
        :type s: str | None

        """
        with self.table_upsert_lock:
            self.table_upsert_sql = s

    def ensure_table(self, cursor):
        """
        Execute on a PSQL connection's cursor the set table upsert command if
        set.

        This method does nothing if no upsert command is set.

        :param cursor: Connection active cursor
        """
        with self.table_upsert_lock:
            if self.table_upsert_sql is not None:
                cursor.execute(self.table_upsert_sql)
                cursor.connection.commit()

    def single_execute(self, cursor_callback, yield_result_rows=False,
                       named=False):
        """
        Perform a single SQL execution in  a new database connection. Handles
        connection/cursor acquisition and release.

        After a successful call to the given callback all result rows are
        yielded, a connection commit is performed in order to finalize any
        changes and to flush the connection (e.g. for pgbouncer).

        Due to this method optionally yielding values, calling this returns a
        generator. This must be iterated over for anything to occur even if
        nothing is to be actively yielded (e.g. execute callback only performs
        writes).

        :param cursor_callback: Function that takes a single positional argument
            that is the active cursor on a new database connection. This
            function should perform any ``execute`` actions on the provided
            cursor required.

            If this callback raises an exception, a rollback is performed on the
            connection and the connection is closed, passing the exception
            forward.
        :type cursor_callback: (psycopg2._psycopg.cursor) -> None

        :param yield_result_rows: Optionally yield rows from each batch
            execution. False by default. If False, nothing is returned.
        :type yield_result_rows: bool

        :param named: If a named cursor should be created, creating a
            server-side cursor. This is only compatible with executions of
            SELECT or VALUES commands.
        :type named: bool

        :return: Iterator over result rows if ``yield_result_rows`` is True.
        :rtype: __generator

        """
        conn = self.get_psql_connection()

        # Optionally create a named cursor to allow server-side iteration. This
        # is required in order to not pull the whole table into memory.
        cursor_name = None
        if named:
            cursor_name = self.get_unique_cursor_name()

        try:
            with conn:
                with conn.cursor() as cur:
                    self.ensure_table(cur)
                with conn.cursor(cursor_name) as cur:
                    # Number of records ``iter(cur)`` must fetch per network
                    # round trip.
                    # - This only maters if the cursor is a named cursor
                    #   (server-side)
                    cur.itersize = self.itersize
                    cursor_callback(cur)

                    # Iterate over returned cursor rows.
                    if yield_result_rows:
                        for r in cur:
                            yield r

                # For server cleaning (e.g. pgbouncer)
                conn.commit()
        except Exception:
            conn.rollback()
            raise
        finally:
            # conn.__exit__ doesn't close connection, just the transaction
            self.connection_pool.putconn(conn)

    def batch_execute(self, iterable, cursor_callback, batch_size,
                      yield_result_rows=False, named=False):
        """
        Batch the given iterable into ``batch_size`` chunks, calling the given
        ``cursor_callback`` with each batch.

        This is generally intended for when the use of the
        ``cursor.executemany`` function is appropriate for statements other than
        SELECT or VALUES.

        When performing a query with a large number of component elements, the
        time between performing the query and getting the first response can be
        long. Additionally, the whole query would need to be constructed first
        which may take a long time or consume a large amount of memory for large
        iterables.

        Splitting an iterable of elements up into batches lessens the
        construction time and memory foot-print of each individual query. This
        in turn improves the response time for individual queries to the server
        as well as keeping the local memory requirements static. A fast response
        time means that results can be iterated out and errors can be revealed
        more quickly.

        This method handles creating a new connection and cursor, ensuring the
        a table's existence (if that query was set) and pulling together batches
        of the given size from the provided opaque iterable.

        After a successful call to the given callback all result rows are
        yielded. After all calls to the given callback are performed, a
        connection commit is performed in order to finalize any changes and to
        flush the connection (e.g. for pgbouncer).

        Due to this method optionally yielding values, calling this returns a
        generator. This must be iterated over for anything to occur even if
        nothing is to be actively yielded.

        :param iterable: Iterable of elements to batch.
        :type iterable: collections.Iterable

        :param cursor_callback: Function that takes two positional arguments:
            the cursor object for the database connection and the current batch
            elements (list-type). The given batch of elements may be of up to
            size ``batch_size`` or less (the "tail" batch). This function should
            perform ``executemany`` actions on the provided cursor for the given
            batch. Results are optionally yielded af this call successfully
            returns.

            If this callback raises an exception, a rollback is performed on the
            connection and the connection is closed, passing the exception
            forward.
        :type cursor_callback: (psycopg2._psycopg.cursor, list) -> None

        :param batch_size: Batch size limit when pulling elements from the input
            iterable. May be a positive integer, 0 or None. A value of 0 or None
            means that no batching occurs and all elements from the iterable are
            collected.

            Once this number of elements are collected from the iterable, the
            callback is called with the collected batch of elements.
        :type batch_size: int | None

        :param yield_result_rows: Optionally yield rows from each batch
            execution. False by default.
        :type yield_result_rows: bool

        :param named: If a named cursor should be created, creating a
            server-side cursor. This is only compatible with executions of
            SELECT or VALUES commands.
        :type named: bool

        :return: Iterator over result rows if ``yield_result_rows`` is True,
            otherwise None.
        :rtype: __generator | None

        """
        if batch_size is None:
            batch_size = 0
        if batch_size < 0:
            raise ValueError("Cannot have a batch size less than 0 "
                             "(given: %s)." % batch_size)

        self._log.debug("starting multi operation (batch_size: %s)", batch_size)

        # Lazy initialize -- only if there are elements to iterate over
        #: :type: None | psycopg2._psycopg.connection
        conn = None

        # Create a named cursor to allow server-side iteration. This is
        # required in order to not pull the whole table into memory.
        cursor_name = None
        if named:
            cursor_name = self.get_unique_cursor_name()

        try:
            batch = []
            i = 0
            for e in iterable:
                if conn is None:
                    conn = self.get_psql_connection()

                batch.append(e)

                if batch_size and len(batch) >= batch_size:
                    i += 1
                    self._log.debug('-- batch %d (size: %d)', i, len(batch))

                    with conn:
                        with conn.cursor() as cur:
                            self.ensure_table(cur)
                        with conn.cursor(cursor_name) as cur:
                            cur.itersize = self.itersize
                            cursor_callback(cur, batch)
                            if yield_result_rows:
                                for r in cur:
                                    yield r
                    batch = []

            if batch:
                self._log.debug('-- tail batch (size: %d)', len(batch))
                with conn:
                    with conn.cursor() as cur:
                        self.ensure_table(cur)
                    with conn.cursor(cursor_name) as cur:
                        cur.itersize = self.itersize
                        cursor_callback(cur, batch)
                        if yield_result_rows:
                            for r in cur:
                                yield r

            if conn is not None:
                conn.commit()
        except Exception:
            if conn is not None:
                conn.rollback()
            raise
        finally:
            # conn.__exit__ doesn't close connection, just the transaction
            if conn is not None:
                self.connection_pool.putconn(conn)
            self._log.debug('-- done')



"""
Tool for training the ITQ functor algorithm's model on descriptors in an
index.

By default, we use all descriptors in the configured index
(``uuids_list_filepath`` is not given a value).

The ``uuids_list_filepath`` configuration property is optional and should
be used to specify a sub-set of descriptors in the configured index to
train on. This only works if the stored descriptors' UUID is a type of
string.
"""

import logging
import os.path

from smqtk.algorithms.nn_index.lsh.functors.itq import ItqFunctor
from smqtk.representation import (
    get_descriptor_index_impls,
)
from smqtk.utils import (
    bin_utils,
    plugin,
)


__author__ = "paul.tunison@kitware.com"


def default_config():
    return {
        "itq_config": ItqFunctor.get_default_config(),
        "uuids_list_filepath": None,
        "descriptor_index": plugin.make_config(get_descriptor_index_impls()),
    }


def cli_parser():
    return bin_utils.basic_cli_parser(__doc__)


def main():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    uuids_list_filepath = config['uuids_list_filepath']

    log.info("Initializing ITQ functor")
    #: :type: smqtk.algorithms.nn_index.lsh.functors.itq.ItqFunctor
    functor = ItqFunctor.from_config(config['itq_config'])

    log.info("Initializing DescriptorIndex [type=%s]",
             config['descriptor_index']['type'])
    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(
        config['descriptor_index'],
        get_descriptor_index_impls(),
    )

    if uuids_list_filepath and os.path.isfile(uuids_list_filepath):
        def uuids_iter():
            with open(uuids_list_filepath) as f:
                for l in f:
                    yield l.strip()
        log.info("Loading UUIDs list from file: %s", uuids_list_filepath)
        d_iter = descriptor_index.get_many_descriptors(uuids_iter())
    else:
        log.info("Using UUIDs from loaded DescriptorIndex (count=%d)",
                 len(descriptor_index))
        d_iter = descriptor_index

    log.info("Fitting ITQ model")
    functor.fit(d_iter)
    log.info("Done")


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

import random
import os.path as osp
import unittest

import numpy as np
from six.moves import range, zip

from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.algorithms import get_nn_index_impls
from smqtk.algorithms.nn_index.mrpt import MRPTNearestNeighborsIndex
from smqtk.exceptions import ReadOnlyError
from smqtk.representation.descriptor_index.memory import MemoryDescriptorIndex


class TestMRPTIndex (unittest.TestCase):

    RAND_SEED = 42

    def _make_inst(self, **kwargs):
        """
        Make an instance of MRPTNearestNeighborsIndex
        """
        if 'random_seed' not in kwargs:
            kwargs.update(random_seed=self.RAND_SEED)
        return MRPTNearestNeighborsIndex(
            MemoryDescriptorIndex(), **kwargs)

    def test_impl_findable(self):
        self.assertIn(MRPTNearestNeighborsIndex.__name__,
                      get_nn_index_impls())

    def test_configuration(self):
        index_filepath = osp.abspath(osp.expanduser('index_filepath'))
        para_filepath = osp.abspath(osp.expanduser('param_fp'))

        # Make configuration based on default
        c = MRPTNearestNeighborsIndex.get_default_config()
        c['index_filepath'] = index_filepath
        c['parameters_filepath'] = para_filepath
        c['descriptor_set']['type'] = 'MemoryDescriptorIndex'

        # Build based on configuration
        index = MRPTNearestNeighborsIndex.from_config(c)
        self.assertEqual(index._index_filepath, index_filepath)
        self.assertEqual(index._index_param_filepath, para_filepath)

        # Test that constructing a new instance from ``index``'s config yields
        # an index with the same configuration (idempotent).
        index2 = MRPTNearestNeighborsIndex.from_config(index.get_config())
        self.assertEqual(index.get_config(), index2.get_config())

    def test_read_only(self):
        v = np.zeros(5, float)
        v[0] = 1.
        d = DescriptorMemoryElement('unit', 0)
        d.set_vector(v)
        test_descriptors = [d]

        index = self._make_inst(read_only=True)
        self.assertRaises(
            ReadOnlyError,
            index.build_index, test_descriptors
        )

    def test_update_index_no_input(self):
        index = self._make_inst()
        self.assertRaises(
            ValueError,
            index.update_index, []
        )

    def test_update_index_new_index(self):
        n = 100
        dim = 8
        d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
        [d.set_vector(np.random.rand(dim)) for d in d_index]

        index = self._make_inst()
        index.update_index(d_index)
        self.assertEqual(index.count(), 100)
        for d in d_index:
            self.assertIn(d, index._descriptor_set)

        # Check that NN can return stuff from the set used.
        # - nearest element to the query element when the query is in the index
        #   should be the query element.
        random.seed(self.RAND_SEED)
        for _ in range(10):
            i = random.randint(0, n-1)
            q = d_index[i]
            n_elems, n_dists = index.nn(q)
            self.assertEqual(n_elems[0], q)

    def test_update_index_additive(self):
        n1 = 100
        n2 = 10
        dim = 8
        set1 = {DescriptorMemoryElement('test', i) for i in range(n1)}
        set2 = {DescriptorMemoryElement('test', i) for i in range(n1, n1+n2)}
        [d.set_vector(np.random.rand(dim)) for d in set1.union(set1 | set2)]

        # Create and build initial index.
        index = self._make_inst()
        index.build_index(set1)
        self.assertEqual(index.count(), len(set1))
        for d in set1:
            self.assertIn(d, index._descriptor_set)

        # Update and check that all intended descriptors are present in index.
        index.update_index(set2)
        set_all = set1 | set2
        self.assertEqual(index.count(), len(set_all))
        for d in set_all:
            self.assertIn(d, index._descriptor_set)

        # Check that NN can return something from the updated set.
        # - nearest element to the query element when the query is in the index
        #   should be the query element.
        for q in set2:
            n_elems, n_dists = index.nn(q)
            self.assertEqual(n_elems[0], q)

    def test_remove_from_index_readonly(self):
        """
        Test that remove causes an error in a readonly instance.
        """
        index = self._make_inst(read_only=True)
        self.assertRaises(
            ReadOnlyError,
            index.remove_from_index, [0]
        )

    def test_remove_from_index_invalid_uid(self):
        """
        Test that error occurs when attempting to remove descriptor UID that
        isn't indexed.
        """
        index = self._make_inst()
        self.assertRaises(
            KeyError,
            index.remove_from_index, [0]
        )

    def test_remove_from_index(self):
        """
        Test expected removal from the index.
        """
        n = 100
        dim = 32
        dset = [DescriptorMemoryElement('test', i) for i in range(n)]
        np.random.seed(self.RAND_SEED)
        [d.set_vector(np.random.rand(dim)) for d in dset]

        index = self._make_inst()
        index.build_index(dset)
        # Test expected initial condition.
        # noinspection PyCompatibility
        self.assertSetEqual(set(index._descriptor_set.iterkeys()),
                            set(range(100)))

        # Try removing some elements.
        d_to_remove = [dset[10], dset[47], dset[82]]
        index.remove_from_index([d.uuid() for d in d_to_remove])

        # Internal descriptor-set should no longer contain the removed
        # descriptor elements.
        for d in d_to_remove:
            self.assertNotIn(d, index._descriptor_set)
        self.assertEqual(len(index._descriptor_set), 97)
        # Make sure that when we query for elements removed, they are not the
        # returned set of things.
        self.assertNotIn(dset[10].uuid(),
                         set(d.uuid() for d in index.nn(dset[10], n)[0]))
        self.assertNotIn(dset[47].uuid(),
                         set(d.uuid() for d in index.nn(dset[10], n)[0]))
        self.assertNotIn(dset[82].uuid(),
                         set(d.uuid() for d in index.nn(dset[10], n)[0]))

    def test_nn_many_descriptors(self):
        np.random.seed(0)

        n = 10 ** 4
        dim = 256
        depth = 5
        num_trees = 10

        d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
        [d.set_vector(np.random.rand(dim)) for d in d_index]
        q = DescriptorMemoryElement('q', -1)
        q.set_vector(np.zeros((dim,)))

        di = MemoryDescriptorIndex()
        mrpt = MRPTNearestNeighborsIndex(
            di, num_trees=num_trees, depth=depth, random_seed=0)
        mrpt.build_index(d_index)

        nbrs, dists = mrpt.nn(q, 10)
        self.assertEqual(len(nbrs), len(dists))
        self.assertEqual(len(nbrs), 10)

    def test_nn_small_leaves(self):
        np.random.seed(0)

        n = 10 ** 4
        dim = 256
        depth = 10
        # L ~ n/2**depth = 10^4 / 2^10 ~ 10
        k = 200
        # 3k/L = 60
        num_trees = 60

        d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
        [d.set_vector(np.random.rand(dim)) for d in d_index]
        q = DescriptorMemoryElement('q', -1)
        q.set_vector(np.zeros((dim,)))

        di = MemoryDescriptorIndex()
        mrpt = MRPTNearestNeighborsIndex(
            di, num_trees=num_trees, depth=depth, random_seed=0)
        mrpt.build_index(d_index)

        nbrs, dists = mrpt.nn(q, k)
        self.assertEqual(len(nbrs), len(dists))
        self.assertEqual(len(nbrs), k)

    def test_nn_pathological_example(self):
        n = 10 ** 4
        dim = 256
        depth = 10
        # L ~ n/2**depth = 10^4 / 2^10 ~ 10
        k = 200
        # 3k/L = 60
        num_trees = 60

        d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
        # Put all descriptors on a line so that different trees get same
        # divisions.
        # noinspection PyTypeChecker
        [d.set_vector(np.full(dim, d.uuid(), dtype=np.float64))
         for d in d_index]
        q = DescriptorMemoryElement('q', -1)
        q.set_vector(np.zeros((dim,)))

        di = MemoryDescriptorIndex()
        mrpt = MRPTNearestNeighborsIndex(
            di, num_trees=num_trees, depth=depth, random_seed=0)
        mrpt.build_index(d_index)

        nbrs, dists = mrpt.nn(q, k)
        self.assertEqual(len(nbrs), len(dists))
        # We should get about 10 descriptors back instead of the requested
        # 200
        self.assertLess(len(nbrs), 20)

    def test_nn_known_descriptors_euclidean_unit(self):
        dim = 5

        ###
        # Unit vectors -- Equal distance
        #
        index = self._make_inst()
        test_descriptors = []
        for i in range(dim):
            v = np.zeros(dim, float)
            v[i] = 1.
            d = DescriptorMemoryElement('unit', i)
            d.set_vector(v)
            test_descriptors.append(d)
        index.build_index(test_descriptors)
        # query descriptor -- zero vector
        # -> all modeled descriptors should be equally distant (unit
        # corners)
        q = DescriptorMemoryElement('query', 0)
        q.set_vector(np.zeros(dim, float))
        r, dists = index.nn(q, n=dim)
        self.assertEqual(len(dists), dim)
        # All dists should be 1.0, r order doesn't matter
        for d in dists:
            self.assertEqual(d, 1.)

    def test_nn_known_descriptors_nearest(self):
        dim = 5

        ###
        # Unit vectors -- Equal distance
        #
        index = self._make_inst()
        test_descriptors = []
        vectors = np.eye(dim, dtype=np.float32)
        for i in range(dim):
            d = DescriptorMemoryElement('unit', i)
            d.set_vector(vectors[i])
            test_descriptors.append(d)
        index.build_index(test_descriptors)
        for i in range(dim):
            # query descriptor -- first point
            q = DescriptorMemoryElement('query', i)
            q.set_vector(vectors[i])
            r, dists = index.nn(q)
            self.assertEqual(len(dists), 1)
            # Distance should be zero (exact match)
            self.assertEqual(dists[0], 0.)
            np.testing.assert_allclose(r[0].vector(), vectors[i])

    def test_nn_known_descriptors_euclidean_ordered(self):
        index = self._make_inst()

        # make vectors to return in a known euclidean distance order
        i = 100
        test_descriptors = []
        for j in range(i):
            d = DescriptorMemoryElement('ordered', j)
            d.set_vector(np.array([j, j*2], float))
            test_descriptors.append(d)
        random.shuffle(test_descriptors)
        index.build_index(test_descriptors)

        # Since descriptors were build in increasing distance from (0,0),
        # returned descriptors for a query of [0,0] should be in index
        # order.
        q = DescriptorMemoryElement('query', 99)
        q.set_vector(np.array([0, 0], float))
        r, dists = index.nn(q, n=i)
        # Because the data is one-dimensional, all of the cells will have
        # the same points (any division will just correspond to a point on
        # the line), and a cell can't have more than half of the points
        self.assertEqual(len(dists), i//2)
        for j, d, dist in zip(range(i), r, dists):
            self.assertEqual(d.uuid(), j)
            np.testing.assert_equal(d.vector(), [j, j*2])



from smqtk.representation import ClassificationElementFactory
from smqtk.representation.classification_element.memory import \
    MemoryClassificationElement


# Default classifier element factory for interfaces.
DFLT_CLASSIFIER_FACTORY = ClassificationElementFactory(
    MemoryClassificationElement, {}
)



from __future__ import division, print_function

import unittest

from smqtk.utils.plugin import (
    make_config,
    to_plugin_config,
    from_plugin_config
)
from smqtk.tests.utils.test_configurable_interface import (
    DummyAlgo1, DummyAlgo2
)


def dummy_getter():
    return {
        'DummyAlgo1': DummyAlgo1,
        'DummyAlgo2': DummyAlgo2,
    }


class TestPluginTools (unittest.TestCase):

    def test_make_config(self):
        self.assertEqual(
            make_config(dummy_getter()),
            {
                'type': None,
                'DummyAlgo1': DummyAlgo1.get_default_config(),
                'DummyAlgo2': DummyAlgo2.get_default_config(),
            }
        )

    def test_to_config(self):
        i = DummyAlgo1()
        d1 = i.get_config()
        c = to_plugin_config(i)
        self.assertEqual(c, {
            'type': 'DummyAlgo1',
            'DummyAlgo1': d1,
        })

        # return should update with updates to
        i.foo = 8
        d2 = i.get_config()
        self.assertNotEqual(d1, d2)
        c = to_plugin_config(i)
        self.assertEqual(c, {
            'type': 'DummyAlgo1',
            'DummyAlgo1': d2,
        })

    def test_from_config(self):
        test_config = {
            'type': 'DummyAlgo1',
            'DummyAlgo1': {'foo': 256, 'bar': 'Some string value'},
            'DummyAlgo2': {
                'child': {'foo': -1, 'bar': 'some other value'},
                'alpha': 1.0,
                'beta': 'euclidean',
            },
            'notAnImpl': {}
        }

        #: :type: DummyAlgo1
        i = from_plugin_config(test_config, dummy_getter())
        self.assertIsInstance(i, DummyAlgo1)
        self.assertEqual(i.foo, 256)
        self.assertEqual(i.bar, 'Some string value')

    def test_from_config_missing_type(self):
        test_config = {
            'DummyAlgo1': {'foo': 256, 'bar': 'Some string value'},
            'DummyAlgo2': {
                'child': {'foo': -1, 'bar': 'some other value'},
                'alpha': 1.0,
                'beta': 'euclidean',
            },
            'notAnImpl': {}
        }
        self.assertRaisesRegexp(
            ValueError,
            "does not have an implementation type specification",
            from_plugin_config,
            test_config, dummy_getter()
        )

    def test_from_config_none_type(self):
        test_config = {
            'type': None,
            'DummyAlgo1': {'foo': 256, 'bar': 'Some string value'},
            'DummyAlgo2': {
                'child': {'foo': -1, 'bar': 'some other value'},
                'alpha': 1.0,
                'beta': 'euclidean',
            },
            'notAnImpl': {}
        }
        self.assertRaisesRegexp(
            ValueError,
            "No implementation type specified",
            from_plugin_config,
            test_config,
            dummy_getter()
        )

    def test_from_config_config_label_mismatch(self):
        test_config = {
            'type': 'not-present-label',
            'DummyAlgo1': {'foo': 256, 'bar': 'Some string value'},
            'DummyAlgo2': {
                'child': {'foo': -1, 'bar': 'some other value'},
                'alpha': 1.0,
                'beta': 'euclidean',
            },
            'notAnImpl': {}
        }
        self.assertRaisesRegexp(
            ValueError,
            "no configuration block was present for that type",
            from_plugin_config,
            test_config,
            dummy_getter()
        )

    def test_from_config_impl_label_mismatch(self):
        test_config = {
            'type': 'notAnImpl',
            'DummyAlgo1': {'foo': 256, 'bar': 'Some string value'},
            'DummyAlgo2': {
                'child': {'foo': -1, 'bar': 'some other value'},
                'alpha': 1.0,
                'beta': 'euclidean',
            },
            'notAnImpl': {}
        }
        self.assertRaisesRegexp(
            ValueError,
            "no plugin implementations are available for that type",
            from_plugin_config,
            test_config,
            dummy_getter()
        )



import random
import unittest
import six

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation.data_element.memory_element import DataMemoryElement


def random_string(length):
    # 32-127 is legible characters
    return ''.join(chr(random.randint(32, 127)) for _ in range(length))


class TestDataMemoryElement (unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.EXPECTED_BYTES = six.b('hello world')
        cls.EXPECTED_CT = 'text/plain'

        cls.VALID_BASE64 = 'aGVsbG8gd29ybGQ='
        cls.INVALID_BASE64 = '$%&^c85swd8a5sw568vs!'

        cls.VALID_B64_URI = 'base64://' + cls.VALID_BASE64
        cls.VALID_DATA_URI = 'data:' + cls.EXPECTED_CT + ';base64,' + \
                             cls.VALID_BASE64

    def test_configuration(self):
        default_config = DataMemoryElement.get_default_config()
        self.assertEqual(default_config,
                         {'bytes': None, 'content_type': None,
                          'readonly': False})

        default_config['bytes'] = 'Hello World.'
        default_config['content_type'] = 'text/plain'
        #: :type: DataMemoryElement
        inst1 = DataMemoryElement.from_config(default_config)
        self.assertEqual(default_config, inst1.get_config())
        self.assertEqual(inst1._bytes, 'Hello World.')
        self.assertEqual(inst1._content_type, 'text/plain')

        inst2 = DataMemoryElement.from_config(inst1.get_config())
        self.assertEqual(inst1, inst2)

    #
    # from_base64 tests
    #

    def test_from_base64_no_ct(self):
        e = DataMemoryElement.from_base64(self.VALID_BASE64)
        self.assertIsInstance(e, DataMemoryElement)
        self.assertEqual(e.get_bytes(), self.EXPECTED_BYTES)

    def test_from_base64_with_ct(self):
        e = DataMemoryElement.from_base64(self.VALID_BASE64, self.EXPECTED_CT)
        self.assertIsInstance(e, DataMemoryElement)
        self.assertEqual(e.get_bytes(), self.EXPECTED_BYTES)
        self.assertEqual(e.content_type(), self.EXPECTED_CT)

    def test_from_base64_null_bytes(self):
        self.assertRaises(
            ValueError,
            DataMemoryElement.from_base64,
            None, None
        )

    def test_from_base64_empty_string(self):
        # Should translate to empty byte string
        e = DataMemoryElement.from_base64('', None)
        self.assertIsInstance(e, DataMemoryElement)
        self.assertEqual(e.get_bytes(), six.b(''))

    #
    # From URI tests
    #

    def test_from_uri_null_uri(self):
        self.assertRaises(
            InvalidUriError,
            DataMemoryElement.from_uri,
            None
        )

    def test_from_uri_empty_string(self):
        # Should return an element with no byte data
        e = DataMemoryElement.from_uri('')
        self.assertIsInstance(e, DataMemoryElement)
        # no base64 data, which should decode to no bytes
        self.assertEqual(e.get_bytes(), six.b(''))

    def test_from_uri_random_string(self):
        rs = random_string(32)
        self.assertRaises(
            InvalidUriError,
            DataMemoryElement.from_uri,
            rs
        )

    def test_from_uri_base64_header_empty_data(self):
        e = DataMemoryElement.from_uri('base64://')
        self.assertIsInstance(e, DataMemoryElement)
        # no base64 data, which should decode to no bytes
        self.assertEqual(e.get_bytes(), six.b(''))

    def test_from_uri_base64_header_invalid_base64(self):
        # URI base64 data contains invalid alphabet characters
        self.assertRaises(
            InvalidUriError,
            DataMemoryElement.from_uri,
            'base64://'+self.INVALID_BASE64
        )

    def test_from_uri_base64_equals_out_of_place(self):
        # '=' characters should only show up at the end of a base64 data string
        self.assertRaises(
            InvalidUriError,
            DataMemoryElement.from_uri,
            'base64://foo=bar'
        )

    def test_from_uri_base64_too_many_equals(self):
        # There should only be a max of 2 '=' characters at the end of the b64
        # data string
        self.assertRaises(
            InvalidUriError,
            DataMemoryElement.from_uri,
            'base64://foobar==='
        )

    def test_from_uri_base64_header(self):
        e = DataMemoryElement.from_uri(self.VALID_B64_URI)
        self.assertIsInstance(e, DataMemoryElement)
        self.assertEqual(e.get_bytes(), self.EXPECTED_BYTES)
        # No content type info in base64 format
        self.assertEqual(e.content_type(), None)

    def test_from_uri_data_format_empty_data(self):
        e = DataMemoryElement.from_uri('data:text/plain;base64,')
        self.assertIsInstance(e, DataMemoryElement)
        # no base64 data, which should decode to no bytes
        self.assertEqual(e.get_bytes(), six.b(''))
        self.assertEqual(e.content_type(), 'text/plain')

    def test_from_uri_data_format_invalid_base64(self):
        self.assertRaises(
            InvalidUriError,
            DataMemoryElement.from_uri,
            'data:text/plain;base64,' + self.INVALID_BASE64
        )

    def test_from_uri_data_format(self):
        e = DataMemoryElement.from_uri(self.VALID_DATA_URI)
        self.assertIsInstance(e, DataMemoryElement)
        self.assertEqual(e.get_bytes(), self.EXPECTED_BYTES)
        self.assertEqual(e.content_type(), self.EXPECTED_CT)

    #
    # Content tests
    #
    def test_is_empty_zero_bytes(self):
        e = DataMemoryElement('')
        self.assertTrue(e.is_empty())

    def test_is_empty_nonzero_bytes(self):
        e = DataMemoryElement('some bytes')
        self.assertFalse(e.is_empty())

    def test_get_bytes_none_bytes(self):
        e = DataMemoryElement()
        self.assertEqual(e.get_bytes(), six.b(''))

    def test_get_bytes_empty_bytes(self):
        e = DataMemoryElement('')
        self.assertEqual(e.get_bytes(), six.b(''))

    def test_get_bytes_some_bytes(self):
        expected_bytes = 'some bytes'
        e = DataMemoryElement(expected_bytes)
        self.assertEqual(e.get_bytes(), expected_bytes)

    def test_writable_default(self):
        v = 'foo'
        e = DataMemoryElement(v)
        self.assertTrue(e.writable())

    def test_writable_when_readonly(self):
        e = DataMemoryElement('', readonly=True)
        self.assertFalse(e.writable())

    def test_writable_when_not_readonly(self):
        e = DataMemoryElement('', readonly=False)
        self.assertTrue(e.writable())

    def test_set_bytes(self):
        bytes_a = six.b('test bytes first set')
        bytes_b = six.b('the second set of bytes')
        e = DataMemoryElement(bytes_a)
        self.assertEqual(e.get_bytes(), bytes_a)
        e.set_bytes(bytes_b)
        self.assertEqual(e.get_bytes(), bytes_b)

    def test_set_bytes_when_readonly(self):
        bytes_a = six.b('test bytes first set')
        bytes_b = six.b('the second set of bytes')
        e = DataMemoryElement(bytes_a, readonly=True)
        self.assertEqual(e.get_bytes(), bytes_a)
        self.assertRaises(
            ReadOnlyError,
            e.set_bytes,
            bytes_b
        )
        self.assertEqual(e.get_bytes(), bytes_a)



import mock
from unittest import TestCase

import smqtk.exceptions
from smqtk.representation.data_element.hbase_element import HBaseDataElement


class TestHBaseDataElement(TestCase):

    DUMMY_CFG = {
        'element_key': 'foobar',
        'binary_column': 'binary_data',
        'hbase_address': 'some_address',
        'hbase_table': 'some_table',
        'timeout': 12345,
    }

    def make_element(self, content):
        e = HBaseDataElement(**self.DUMMY_CFG)
        # Pretend that the implementation is actually available and mock out
        # dependency functionality.
        e.content_type = mock.MagicMock()
        e._new_hbase_table_connection = mock.MagicMock()
        e._new_hbase_table_connection().row.return_value = {
            self.DUMMY_CFG['binary_column']: content
        }
        return e

    @classmethod
    def setUpClass(cls):
        # Pretend that the implementation is actually available and mock out
        # dependency functionality.
        HBaseDataElement.is_usable = mock.MagicMock(return_value=True)

    def test_config(self):
        cfg = HBaseDataElement.get_default_config()
        self.assertEqual(cfg, {
            'element_key': None,
            'binary_column': None,
            'hbase_address': None,
            'hbase_table': None,
            'timeout': 10000,
        })

        cfg = self.DUMMY_CFG
        #: :type: HBaseDataElement
        e = HBaseDataElement.from_config(cfg)

        self.assertEqual(e.element_key, cfg['element_key'])
        self.assertEqual(e.binary_column, cfg['binary_column'])
        self.assertEqual(e.hbase_address, cfg['hbase_address'])
        self.assertEqual(e.hbase_table, cfg['hbase_table'])
        self.assertEqual(e.timeout, cfg['timeout'])

        # output should be the same as what we constructed with in this case
        e_get_cfg = e.get_config()
        self.assertEqual(e_get_cfg, cfg)

    def test_get_bytes(self):
        expected_bytes = 'foo bar test string'
        e = self.make_element(expected_bytes)
        self.assertEqual(e.get_bytes(), expected_bytes)

    def test_is_empty_zero_bytes(self):
        # Simulate empty bytes
        e = self.make_element('')
        self.assertTrue(e.is_empty())

    def test_is_empty_nonzero_bytes(self):
        # Simulate non-empty bytes
        e = self.make_element('some bytes')
        self.assertFalse(e.is_empty())

    def test_writable(self):
        # Read-only element
        e = self.make_element('')
        self.assertFalse(e.writable())

    def test_set_bytes(self):
        # Read-only element
        e = self.make_element('')
        self.assertRaises(
            smqtk.exceptions.ReadOnlyError,
            e.set_bytes, 'some bytes'
        )



from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import random
import unittest

import numpy as np
import six
from six.moves import range, zip

from smqtk.algorithms import get_nn_index_impls
from smqtk.algorithms.nn_index.faiss import FaissNearestNeighborsIndex
from smqtk.exceptions import ReadOnlyError
from smqtk.representation.data_element.memory_element import (
    DataMemoryElement,
)
from smqtk.representation.descriptor_element.local_elements import (
    DescriptorMemoryElement,
)
from smqtk.representation.descriptor_index.memory import MemoryDescriptorIndex
from smqtk.representation.key_value.memory import MemoryKeyValueStore

if FaissNearestNeighborsIndex.is_usable():

    class TestFAISSIndex (unittest.TestCase):

        RAND_SEED = 42

        def _make_inst(self, descriptor_set=None, idx2uid_kvs=None,
                       uid2idx_kvs=None, **kwargs):
            """
            Make an instance of FaissNearestNeighborsIndex
            """
            if 'random_seed' not in kwargs:
                kwargs.update(random_seed=self.RAND_SEED)
            if descriptor_set is None:
                descriptor_set = MemoryDescriptorIndex()
            if idx2uid_kvs is None:
                idx2uid_kvs = MemoryKeyValueStore()
            if uid2idx_kvs is None:
                uid2idx_kvs = MemoryKeyValueStore()
            return FaissNearestNeighborsIndex(descriptor_set, idx2uid_kvs,
                                              uid2idx_kvs, **kwargs)

        def test_impl_findable(self):
            self.assertIn(FaissNearestNeighborsIndex.__name__,
                          get_nn_index_impls())

        def test_configuration(self):
            # Make configuration based on default
            c = FaissNearestNeighborsIndex.get_default_config()

            self.assertIn('MemoryDescriptorIndex', c['descriptor_set'])
            c['descriptor_set']['type'] = 'MemoryDescriptorIndex'

            self.assertIn('MemoryKeyValueStore', c['idx2uid_kvs'])
            c['idx2uid_kvs']['type'] = 'MemoryKeyValueStore'

            self.assertIn('MemoryKeyValueStore', c['uid2idx_kvs'])
            c['uid2idx_kvs']['type'] = 'MemoryKeyValueStore'

            self.assertIn('DataMemoryElement', c['index_element'])
            c['index_element']['type'] = 'DataMemoryElement'

            self.assertIn('DataMemoryElement', c['index_param_element'])
            c['index_param_element']['type'] = 'DataMemoryElement'

            # # Build based on configuration
            index = FaissNearestNeighborsIndex.from_config(c)
            self.assertEqual(index.factory_string, 'IVF1,Flat')
            self.assertIsInstance(index.factory_string, six.string_types)

            # Test that constructing a new instance from ``index``'s config
            # yields an index with the same configuration (idempotent).
            index2 = FaissNearestNeighborsIndex.from_config(
                index.get_config())
            self.assertEqual(index.get_config(), index2.get_config())

        def test_configuration_null_persistence(self):
            # Make configuration based on default
            c = FaissNearestNeighborsIndex.get_default_config()
            c['descriptor_set']['type'] = 'MemoryDescriptorIndex'
            c['idx2uid_kvs']['type'] = 'MemoryKeyValueStore'
            c['uid2idx_kvs']['type'] = 'MemoryKeyValueStore'

            # # Build based on configuration
            index = FaissNearestNeighborsIndex.from_config(c)
            self.assertEqual(index.factory_string, 'IVF1,Flat')
            self.assertIsInstance(index.factory_string, six.string_types)

            # Test that constructing a new instance from ``index``'s config
            # yields an index with the same configuration (idempotent).
            index2 = FaissNearestNeighborsIndex.from_config(
                index.get_config())
            self.assertEqual(index.get_config(), index2.get_config())

        def test_build_index_read_only(self):
            v = np.zeros(5, float)
            v[0] = 1.
            d = DescriptorMemoryElement('unit', 0)
            d.set_vector(v)
            test_descriptors = [d]

            index = self._make_inst(read_only=True)
            self.assertRaises(
                ReadOnlyError,
                index.build_index, test_descriptors
            )

        def test_update_index_no_input(self):
            index = self._make_inst()
            self.assertRaises(
                ValueError,
                index.update_index, []
            )

        def test_update_index_new_index(self):
            n = 100
            dim = 8
            d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
            [d.set_vector(np.random.rand(dim)) for d in d_index]

            index = self._make_inst()
            index.update_index(d_index)
            self.assertEqual(index.count(), 100)
            for d in d_index:
                self.assertIn(d, index._descriptor_set)

            # Check that NN can return stuff from the set used.
            # - nearest element to the query element when the query is in the
            #   index should be the query element.
            random.seed(self.RAND_SEED)
            for _ in range(10):
                i = random.randint(0, n-1)
                q = d_index[i]
                n_elems, n_dists = index.nn(q)
                self.assertEqual(n_elems[0], q)

        def test_update_index_additive(self):
            n1 = 100
            n2 = 10
            dim = 8
            set1 = {DescriptorMemoryElement('test', i) for i in range(n1)}
            set2 = {DescriptorMemoryElement('test', i)
                    for i in range(n1, n1+n2)}
            [d.set_vector(np.random.rand(dim)) for d in (set1 | set2)]

            # Create and build initial index.
            index = self._make_inst()
            index.build_index(set1)
            self.assertEqual(index.count(), len(set1))
            for d in set1:
                self.assertIn(d, index._descriptor_set)

            # Update and check that all intended descriptors are present in
            # index.
            index.update_index(set2)
            set_all = set1 | set2
            self.assertEqual(index.count(), len(set_all))
            for d in set_all:
                self.assertIn(d, index._descriptor_set)

            # Check that NN can return something from the updated set.
            # - nearest element to the query element when the query is in the
            #   index should be the query element.
            for q in set_all:
                n_elems, n_dists = index.nn(q)
                self.assertEqual(n_elems[0], q)

        def test_persistence_with_update_index(self):
            n1 = 100
            n2 = 10
            dim = 8
            set1 = {DescriptorMemoryElement('test', i) for i in range(n1)}
            set2 = {DescriptorMemoryElement('test', i)
                    for i in range(n1, n1+n2)}
            [d.set_vector(np.random.rand(dim)) for d in (set1 | set2)]

            # Create index with persistent entities
            index_element = DataMemoryElement(
                content_type='application/octet-stream')
            index_param_element = DataMemoryElement(
                content_type='text/plain')
            index = self._make_inst(
                index_element=index_element,
                index_param_element=index_param_element)
            descriptor_set = index._descriptor_set
            idx2uid_kvs = index._idx2uid_kvs
            uid2idx_kvs = index._uid2idx_kvs

            # Build initial index.
            index.build_index(set1)
            self.assertEqual(index.count(), len(set1))
            for d in set1:
                self.assertIn(d, index._descriptor_set)

            # Update and check that all intended descriptors are present in
            # index.
            index.update_index(set2)
            set_all = set1 | set2
            self.assertEqual(index.count(), len(set_all))
            for d in set_all:
                self.assertIn(d, index._descriptor_set)

            del index
            index = self._make_inst(
                descriptor_set=descriptor_set,
                idx2uid_kvs=idx2uid_kvs,
                uid2idx_kvs=uid2idx_kvs,
                index_element=index_element,
                index_param_element=index_param_element)

            # Check that NN can return something from the updated set.
            # - nearest element to the query element when the query is in the
            #   index should be the query element.
            for q in set_all:
                n_elems, n_dists = index.nn(q)
                self.assertEqual(n_elems[0], q)

        def test_remove_from_index_readonly(self):
            """
            Test that we cannot call remove when the instance is read-only.
            """
            index = self._make_inst(read_only=True)
            self.assertRaises(
                ReadOnlyError,
                index.remove_from_index, [0]
            )

        def test_remove_from_index_keyerror_empty_index(self):
            """
            Test that any key should cause a key error on an empty index.
            """
            index = self._make_inst()
            self.assertRaisesRegexp(
                KeyError, '0',
                index.remove_from_index, [0]
            )
            self.assertRaisesRegexp(
                KeyError, '0',
                index.remove_from_index, ['0']
            )
            # Only includes the first key that's erroneous in the KeyError inst
            self.assertRaisesRegexp(
                KeyError, '0',
                index.remove_from_index, [0, 'other']
            )

        def test_remove_from_index_keyerror(self):
            """
            Test that we do not impact the index by trying to remove an invalid
            key.
            """
            n = 100
            dim = 8
            dset = {DescriptorMemoryElement('test', i) for i in range(n)}
            [d.set_vector(np.random.rand(dim)) for d in dset]

            index = self._make_inst()
            index.build_index(dset)

            # Try removing 2 invalid entries
            self.assertRaises(
                KeyError,
                index.remove_from_index, [100, 'something']
            )

            # Make sure that all indexed descriptors correctly return
            # themselves from an NN call.
            for d in dset:
                self.assertEqual(index.nn(d, 1)[0][0], d)

        def test_remove_from_index(self):
            """
            Test that we can actually remove from the index.
            """
            n = 100
            dim = 8
            dset = {DescriptorMemoryElement('test', i) for i in range(n)}
            [d.set_vector(np.random.rand(dim)) for d in dset]

            index = self._make_inst()
            index.build_index(dset)

            # Try removing two valid descriptors
            uids_to_remove = [10, 98]
            index.remove_from_index(uids_to_remove)

            # Check that every other element is still in the index.
            self.assertEqual(len(index), 98)
            for d in dset:
                if d.uuid() not in uids_to_remove:
                    self.assertEqual(index.nn(d, 1)[0][0], d)

            # Check that descriptors matching removed uids cannot be queried
            # out of the index.
            for d in dset:
                if d.uuid() in uids_to_remove:
                    self.assertNotEqual(index.nn(d, 1)[0][0], d)

        def test_remove_then_add(self):
            """
            Test that we can remove from the index and then add to it again.
            """
            n1 = 100
            n2 = 10
            dim = 8
            set1 = [DescriptorMemoryElement('test', i) for i in range(n1)]
            set2 = [DescriptorMemoryElement('test', i)
                    for i in range(n1, n1 + n2)]
            [d.set_vector(np.random.rand(dim)) for d in (set1 + set2)]
            uids_to_remove = [10, 98]

            index = self._make_inst()
            index.build_index(set1)
            index.remove_from_index(uids_to_remove)
            index.update_index(set2)

            self.assertEqual(len(index), 108)
            # Removed descriptors should not be in return queries.
            self.assertNotEqual(index.nn(set1[10], 1)[0][0], set1[10])
            self.assertNotEqual(index.nn(set1[98], 1)[0][0], set1[98])
            # Every other descriptor should be queryable
            for d in set1 + set2:
                if d.uuid() not in uids_to_remove:
                    self.assertEqual(index.nn(d, 1)[0][0], d)
            self.assertEqual(index._next_index, 110)

        def test_nn_many_descriptors(self):
            np.random.seed(0)

            n = 10 ** 4
            dim = 256

            d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
            [d.set_vector(np.random.rand(dim)) for d in d_index]
            q = DescriptorMemoryElement('q', -1)
            q.set_vector(np.zeros((dim,)))

            faiss_index = self._make_inst()
            faiss_index.build_index(d_index)

            nbrs, dists = faiss_index.nn(q, 10)
            self.assertEqual(len(nbrs), len(dists))
            self.assertEqual(len(nbrs), 10)

        def test_nn_non_flat_index(self):
            faiss_index = self._make_inst(factory_string='IVF256,Flat')
            self.assertEqual(faiss_index.factory_string, 'IVF256,Flat')

            np.random.seed(self.RAND_SEED)
            n = 10 ** 4
            dim = 256

            d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
            [d.set_vector(np.random.rand(dim)) for d in d_index]
            q = DescriptorMemoryElement('q', -1)
            q.set_vector(np.zeros((dim,)))

            faiss_index.build_index(d_index)

            nbrs, dists = faiss_index.nn(q, 10)
            self.assertEqual(len(nbrs), len(dists))
            self.assertEqual(len(nbrs), 10)

        def test_nn_preprocess_index(self):
            faiss_index = self._make_inst(factory_string='PCAR64,IVF1,Flat')
            self.assertEqual(faiss_index.factory_string, 'PCAR64,IVF1,Flat')

            np.random.seed(self.RAND_SEED)
            n = 10 ** 4
            dim = 256

            d_index = [DescriptorMemoryElement('test', i) for i in range(n)]
            [d.set_vector(np.random.rand(dim)) for d in d_index]
            q = DescriptorMemoryElement('q', -1)
            q.set_vector(np.zeros((dim,)))

            faiss_index.build_index(d_index)

            nbrs, dists = faiss_index.nn(q, 10)
            self.assertEqual(len(nbrs), len(dists))
            self.assertEqual(len(nbrs), 10)

        def test_nn_known_descriptors_euclidean_unit(self):
            dim = 5

            ###
            # Unit vectors -- Equal distance
            #
            index = self._make_inst()
            test_descriptors = []
            for i in range(dim):
                v = np.zeros(dim, float)
                v[i] = 1.
                d = DescriptorMemoryElement('unit', i)
                d.set_vector(v)
                test_descriptors.append(d)
            index.build_index(test_descriptors)
            # query descriptor -- zero vector
            # -> all modeled descriptors should be equally distant (unit
            # corners)
            q = DescriptorMemoryElement('query', 0)
            q.set_vector(np.zeros(dim, float))
            r, dists = index.nn(q, n=dim)
            self.assertEqual(len(dists), dim)
            # All dists should be 1.0, r order doesn't matter
            for d in dists:
                self.assertEqual(d, 1.)

        def test_nn_known_descriptors_nearest(self):
            dim = 5

            ###
            # Unit vectors -- Equal distance
            #
            index = self._make_inst()
            test_descriptors = []
            vectors = np.eye(dim, dtype=np.float32)
            for i in range(dim):
                d = DescriptorMemoryElement('unit', i)
                d.set_vector(vectors[i])
                test_descriptors.append(d)
            index.build_index(test_descriptors)
            # query descriptor -- first point
            q = DescriptorMemoryElement('query', 0)
            q.set_vector(vectors[0])
            r, dists = index.nn(q)
            self.assertEqual(len(dists), 1)
            # Distance should be zero
            self.assertEqual(dists[0], 0.)
            self.assertItemsEqual(r[0].vector(), vectors[0])

        def test_nn_known_descriptors_euclidean_ordered(self):
            index = self._make_inst()

            # make vectors to return in a known euclidean distance order
            i = 100
            test_descriptors = []
            for j in range(i):
                d = DescriptorMemoryElement('ordered', j)
                d.set_vector(np.array([j, j*2], float))
                test_descriptors.append(d)
            random.shuffle(test_descriptors)
            index.build_index(test_descriptors)

            # Since descriptors were built in increasing distance from (0,0),
            # returned descriptors for a query of [0,0] should be in index
            # order.
            q = DescriptorMemoryElement('query', 99)
            q.set_vector(np.array([0, 0], float))
            r, dists = index.nn(q, n=i)

            self.assertEqual(len(dists), i)
            for j, d, dist in zip(range(i), r, dists):
                self.assertEqual(d.uuid(), j)
                np.testing.assert_equal(d.vector(), [j, j*2])



import time

from smqtk.utils import SmqtkObject


class SimpleTimer (SmqtkObject):
    """
    Little class to wrap the timing of things. To be use with the ``with``
    statement.
    """

    def __init__(self, msg, log_func=None, *args):
        """
        Additional arguments are passed to the logging method
        :param msg:
        :param log_func:
        :param args:
        :return:
        """
        self._log_func = log_func
        self._msg = msg
        self._msg_args = args
        self._s = 0.0

    def __enter__(self):
        if self._log_func:
            self._log_func(self._msg, *self._msg_args)
        else:
            self._log.info(self._msg % self._msg_args)
        self._s = time.time()

    def __exit__(self, *_):
        if self._log_func:
            self._log_func("%s -> %f s", self._msg % self._msg_args,
                           time.time() - self._s)
        else:
            self._log.info("%s -> %f s" % (self._msg % self._msg_args,
                                           time.time() - self._s))



import base64
import json
import mock
import os
import unittest

from smqtk.representation import DescriptorElement
from smqtk.web.iqr_service import IqrService

from smqtk.tests.web.iqr_service.stubs import \
    STUB_MODULE_PATH, \
    StubDescriptorIndex, StubDescrGenerator, StubNearestNeighborIndex


class TestIqrService (unittest.TestCase):

    # Patch in this module for stub implementation access.
    # noinspection PyUnresolvedReferences
    @mock.patch.dict(os.environ, {
        "CLASSIFIER_PATH": STUB_MODULE_PATH,
        "DESCRIPTOR_INDEX_PATH": STUB_MODULE_PATH,
        "DESCRIPTOR_GENERATOR_PATH": STUB_MODULE_PATH,
        "NN_INDEX_PATH": STUB_MODULE_PATH,
    })
    def setUp(self):
        """
        Make an instance of the IqrService flask application with stub
        algorithms.
        """
        # Setup configuration for test application
        config = IqrService.get_default_config()
        plugin_config = config['iqr_service']['plugins']

        # Use basic in-memory representation types.
        plugin_config['classification_factory']['type'] =\
            'MemoryClassificationElement'
        plugin_config['descriptor_factory']['type'] = \
            'DescriptorMemoryElement'
        plugin_config['descriptor_index']['type'] = 'StubDescriptorIndex'

        # Set up dummy algorithm types
        plugin_config['classifier_config']['type'] = 'StubClassifier'
        plugin_config['descriptor_generator']['type'] = 'StubDescrGenerator'
        plugin_config['neighbor_index']['type'] = 'StubNearestNeighborIndex'

        self.app = IqrService(config)

    def assertStatusCode(self, r, code):
        """
        :type r: :type: flask.wrappers.Response
        :type code: int
        """
        self.assertEqual(r.status_code, code)

    def assertJsonMessageRegex(self, r, regex):
        """
        :type r: flask.wrappers.Response
        :type regex: str
        """
        self.assertRegexpMatches(json.loads(r.data.decode())['message'], regex)

    # Test Methods ############################################################

    def test_is_ready(self):
        # Test that the is_ready endpoint returns the expected values.
        #: :type: flask.wrappers.Response
        r = self.app.test_client().get('/is_ready')
        self.assertStatusCode(r, 200)
        self.assertJsonMessageRegex(r, "Yes, I'm alive.")
        self.assertIsInstance(self.app.descriptor_index, StubDescriptorIndex)
        self.assertIsInstance(self.app.descriptor_generator, StubDescrGenerator)
        self.assertIsInstance(self.app.neighbor_index, StubNearestNeighborIndex)

    def test_add_descriptor_from_data_no_args(self):
        """ Test that providing no arguments causes a 400 error. """
        r = self.app.test_client().post('/add_descriptor_from_data')
        self.assertStatusCode(r, 400)

    def test_add_descriptor_from_data_no_b64(self):
        """ Test that providing no b64 data is caught + errors. """
        query_data = {
            "content_type": "text/plain"
        }
        r = self.app.test_client().post('/add_descriptor_from_data',
                                        data=query_data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "No or empty base64 data")

    def test_add_descriptor_from_data_no_content_type(self):
        query_data = {
            'data_b64': base64.b64encode(b"some test data").decode()
        }
        r = self.app.test_client().post('/add_descriptor_from_data',
                                        data=query_data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "No data mimetype")

    def test_add_descriptor_from_data_invalid_base64(self):
        """ Test sending some bad base64 data. """
        query_data = {
            'data_b64': 'not valid b64',
            'content_type': 'text/plain',
        }
        r = self.app.test_client().post('/add_descriptor_from_data',
                                        data=query_data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Failed to parse base64 data")

    def test_add_descriptor_from_data(self):
        expected_bytes = b"some test bytes"
        expected_base64 = base64.b64encode(expected_bytes).decode()
        expected_contenttype = 'text/plain'
        expected_descriptor_uid = 'test-descr-uid'
        expected_descriptor = mock.MagicMock(spec=DescriptorElement)
        expected_descriptor.uuid.return_value = expected_descriptor_uid

        self.app.describe_base64_data = mock.Mock(
            return_value=expected_descriptor
        )
        self.app.descriptor_index.add_descriptor = mock.Mock()

        query_data = {
            "data_b64": expected_base64,
            "content_type": expected_contenttype,
        }
        r = self.app.test_client().post('/add_descriptor_from_data',
                                        data=query_data)

        self.app.describe_base64_data.assert_called_once_with(
            expected_base64, expected_contenttype
        )
        self.app.descriptor_index.add_descriptor.assert_called_once_with(
            expected_descriptor
        )

        self.assertStatusCode(r, 201)
        self.assertJsonMessageRegex(r, "Success")
        r_json = json.loads(r.data.decode())
        self.assertEqual(r_json['uid'], expected_descriptor_uid)

    def test_get_nn_index_status(self):
        self.app.neighbor_index.count = mock.Mock(return_value=0)
        with self.app.test_client() as tc:
            r = tc.get('/nn_index')
            self.assertStatusCode(r, 200)
            self.assertJsonMessageRegex(r, "Success")
            self.assertEqual(json.loads(r.data.decode())['index_size'], 0)

        self.app.neighbor_index.count = mock.Mock(return_value=89756234876)
        with self.app.test_client() as tc:
            r = tc.get('/nn_index')
            self.assertStatusCode(r, 200)
            self.assertJsonMessageRegex(r, "Success")
            self.assertEqual(json.loads(r.data.decode())['index_size'], 89756234876)

    def test_update_nn_index_no_args(self):
        """ Test that error is returned when no arguments provided """
        r = self.app.test_client().post('/nn_index')
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "No descriptor UID JSON provided")

    def test_update_nn_index_bad_json_parse_error(self):
        """ Test handling a bad json string """
        r = self.app.test_client().post('/nn_index',
                                        data=dict(
                                            descriptor_uids='not-valid-json'
                                        ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "JSON parsing error")

    def test_update_nn_index_bad_json_not_a_list(self):
        r = self.app.test_client().post('/nn_index',
                                        data=dict(
                                            descriptor_uids='6.2'
                                        ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "JSON provided is not a list")

    def test_update_nn_index_bad_json_empty_list(self):
        r = self.app.test_client().post('/nn_index',
                                        data=dict(
                                            descriptor_uids='[]'
                                        ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "JSON list is empty")

    def test_update_nn_index_bad_json_invalid_parts(self):
        """ All UIDs provided should be hashable values. """
        r = self.app.test_client().post('/nn_index',
                                        data=dict(
                                            descriptor_uids='["a", 4, []]'
                                        ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Not all JSON list parts were hashable "
                                       "values.")

    def test_update_nn_index_uid_not_a_key(self):
        """
        Test that providing one or more UIDs that are not a part of the index
        yields an error.
        """
        def key_error_raise(*_, **__):
            raise KeyError('test-key')

        self.app.descriptor_index.get_many_descriptors = mock.Mock(
            side_effect=key_error_raise
        )
        # Pretend any UID except 2 and "hello" are not contained.
        self.app.descriptor_index.has_descriptor = mock.Mock(
            side_effect=lambda k: k == 2 or k == "hello"
        )

        expected_list = [0, 1, 2, 'hello', 'foobar']
        expected_list_json = '[0, 1, 2, "hello", "foobar"]'
        expected_bad_uids = [0, 1, 'foobar']
        r = self.app.test_client().post('/nn_index',
                                        data=dict(
                                            descriptor_uids=expected_list_json,
                                        ))
        self.app.descriptor_index.get_many_descriptors.assert_called_once_with(
            expected_list
        )
        self.app.descriptor_index.has_descriptor.assert_any_call(0)
        self.app.descriptor_index.has_descriptor.assert_any_call(1)
        self.app.descriptor_index.has_descriptor.assert_any_call(2)
        self.app.descriptor_index.has_descriptor.assert_any_call("hello")
        self.app.descriptor_index.has_descriptor.assert_any_call("foobar")
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Some provided UIDs do not exist in "
                                       "the current index")
        r_json = json.loads(r.data.decode())
        self.assertListEqual(r_json['bad_uids'], expected_bad_uids)

    def test_update_nn_index_delayed_key_error(self):
        """
        Some DescriptorIndex implementations of get_many_descriptors use the
        yield statement and thus won't potentially generate key errors until
        actually iterated within the update call.  Test that this is correctly
        caught.
        """
        expected_error_key = 'expected-error-key'

        def generator_key_error(*_, **__):
            raise KeyError(expected_error_key)
            # make a generator
            # noinspection PyUnreachableCode
            yield

        self.app.descriptor_index.get_many_descriptors = mock.Mock(
            side_effect=generator_key_error
        )
        # Pretend any UID except 2 and "hello" are not contained.
        self.app.descriptor_index.has_descriptor = mock.Mock(
            side_effect=lambda k: k == 2 or k == "hello"
        )

        expected_list = [0, 1, 2, 'hello', 'foobar']
        expected_list_json = '[0, 1, 2, "hello", "foobar"]'
        expected_bad_uids = [0, 1, 'foobar']
        r = self.app.test_client().post('/nn_index',
                                        data=dict(
                                            descriptor_uids=expected_list_json,
                                        ))
        self.app.descriptor_index.get_many_descriptors.assert_called_once_with(
            expected_list
        )
        self.app.descriptor_index.has_descriptor.assert_any_call(0)
        self.app.descriptor_index.has_descriptor.assert_any_call(1)
        self.app.descriptor_index.has_descriptor.assert_any_call(2)
        self.app.descriptor_index.has_descriptor.assert_any_call("hello")
        self.app.descriptor_index.has_descriptor.assert_any_call("foobar")
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Some provided UIDs do not exist in "
                                       "the current index")
        r_json = json.loads(r.data.decode())
        self.assertListEqual(r_json['bad_uids'], expected_bad_uids)

    def test_update_nn_index(self):
        expected_descriptors = [
            mock.Mock(spec=DescriptorElement),
            mock.Mock(spec=DescriptorElement),
            mock.Mock(spec=DescriptorElement),
        ]
        expected_uid_list = [0, 1, 2]
        expected_uid_list_json = json.dumps(expected_uid_list)
        expected_new_index_count = 10

        self.app.descriptor_index.get_many_descriptors = mock.Mock(
            return_value=expected_descriptors
        )
        self.app.neighbor_index.update_index = mock.Mock()
        self.app.neighbor_index.count = mock.Mock(
            return_value=expected_new_index_count
        )

        data = dict(
            descriptor_uids=expected_uid_list_json
        )
        r = self.app.test_client().post('/nn_index', data=data)

        self.app.descriptor_index.get_many_descriptors.assert_called_once_with(
            expected_uid_list
        )
        self.app.neighbor_index.update_index.assert_called_once_with(
            expected_descriptors
        )

        self.assertStatusCode(r, 200)
        r_json = json.loads(r.data.decode())
        self.assertListEqual(r_json['descriptor_uids'], expected_uid_list)
        self.assertEqual(r_json['index_size'], expected_new_index_count)

    def test_remove_from_nn_index_no_uids(self):
        """
        Test that error is properly returned when failing to pass any UIDs to
        the endpoint.
        """
        with self.app.test_client() as tc:
            r = tc.delete('/nn_index')
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, "No descriptor UID JSON provided")

    def test_remove_from_nn_index_bad_json_parse_error(self):
        """
        Test expected error from had JSON input.
        """
        with self.app.test_client() as tc:
            r = tc.delete('/nn_index',
                          data={'descriptor_uids': 'not-valid-json'})
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, "JSON parsing error: ")

    def test_remove_from_nn_index_bad_json_not_a_list(self):
        """
        Test expected error from not sending a list.
        """
        with self.app.test_client() as tc:
            r = tc.delete('/nn_index', data={'descriptor_uids': 6.2})
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, 'JSON provided is not a list.')

    def test_remove_from_nn_index_bad_json_list_empty(self):
        """
        Test expected error when passed an empty JSON list.
        """
        with self.app.test_client() as tc:
            r = tc.delete('/nn_index',
                          data={'descriptor_uids': '[]'})
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, "JSON list is empty")

    def test_remove_from_nn_index_bad_json_invalid_parts(self):
        """
        Test expected error when input JSON contains non-hashable values.
        """
        with self.app.test_client() as tc:
            r = tc.delete('/nn_index',
                          data=dict(
                              descriptor_uids='["a", 1, []]'
                          ))
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, 'Not all JSON list parts were '
                                           'hashable values\.')

    def test_remove_from_nn_index_uid_not_a_key(self):
        """
        Test expected error when providing one or more descriptor UIDs not
        currently indexed.
        """
        expected_exception_msg = 'expected KeyError value'

        def raiseKeyError(*_, **__):
            raise KeyError(expected_exception_msg)

        # Simulate a KeyError being raised by nn-index remove call.
        self.app.neighbor_index.remove_from_index = mock.Mock(
            side_effect=raiseKeyError
        )

        with self.app.test_client() as tc:
            r = tc.delete('/nn_index',
                          data={
                              'descriptor_uids': '[0, 1, 2]'
                          })
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, "Some provided UIDs do not exist "
                                           "in the current index")
            r_json = json.loads(r.data.decode())
            self.assertListEqual(r_json['bad_uids'], [expected_exception_msg])

    def test_remove_from_nn_index(self):
        """
        Test we can remove from the nn-index via the endpoint.
        """
        self.app.neighbor_index.remove_from_index = mock.Mock()
        self.app.neighbor_index.count = mock.Mock(return_value=3)

        with self.app.test_client() as tc:
            r = tc.delete('/nn_index',
                          data=dict(
                              descriptor_uids='[1, 2, 3]'
                          ))

            self.app.neighbor_index.remove_from_index.assert_called_once_with(
                [1, 2, 3]
            )

            self.assertStatusCode(r, 200)
            self.assertJsonMessageRegex(r, "Success")
            r_json = json.loads(r.data.decode())
            self.assertListEqual(r_json['descriptor_uids'], [1, 2, 3])
            self.assertEqual(r_json['index_size'], 3)

    def test_data_nearest_neighbors_no_base64(self):
        """ Test not providing base64 """
        data = dict(
            # data_b64=base64.b64encode('test-data'),
            content_type='text/plain',
            k=10,
        )
        r = self.app.test_client().post('/data_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, 'No or empty base64 data provided')

    def test_data_nearest_neighbors_no_contenttype(self):
        """ Test not providing base64 """
        data = dict(
            data_b64=base64.b64encode(b'test-data').decode(),
            # content_type='text/plain',
            k=10,
        )
        r = self.app.test_client().post('/data_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, 'No data mimetype provided')

    def test_data_nearest_neighbors_no_k(self):
        """ Test not providing base64 """
        data = dict(
            data_b64=base64.b64encode(b'test-data').decode(),
            content_type='text/plain',
            # k=10,
        )
        r = self.app.test_client().post('/data_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "No 'k' value provided")

    def test_data_nearest_neighbors_bad_k_json(self):
        """ Test string for k """
        data = dict(
            data_b64=base64.b64encode(b'test-data').decode(),
            content_type='text/plain',
            k="10.2",  # float string fails an int cast.
        )
        r = self.app.test_client().post('/data_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Failed to convert 'k' argument to an "
                                       "integer")

    def test_data_nearest_neighbors_bad_base64(self):
        data = dict(
            data_b64="not-valid-base-64%",
            content_type='text/plain',
            k=10,
        )
        r = self.app.test_client().post('/data_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Failed to parse base64 data")

    def test_data_nearest_neighbors(self):
        expected_uids = ['a', 'b', 'c']
        expected_neighbors = []
        for uid in expected_uids:
            e = mock.Mock(spec=DescriptorElement)
            e.uuid.return_value = uid
            expected_neighbors.append(e)
        expected_dists = [1, 2, 3]

        self.app.describe_base64_data = mock.Mock()
        self.app.neighbor_index.nn = mock.Mock(
            return_value=[expected_neighbors, expected_dists]
        )

        data = dict(
            data_b64=base64.b64encode(b'test-data').decode(),
            content_type='text/plain',
            k=3,  # float string fails an int cast.
        )
        r = self.app.test_client().post('/data_nearest_neighbors',
                                        data=data)
        self.app.describe_base64_data.assert_called_once_with(
            data['data_b64'], data['content_type']
        )
        # noinspection PyArgumentList
        self.app.neighbor_index.nn.assert_called_once_with(
            self.app.describe_base64_data(), data['k']
        )

        self.assertStatusCode(r, 200)
        r_json = json.loads(r.data.decode())
        self.assertListEqual(r_json['neighbor_uids'], expected_uids)
        self.assertListEqual(r_json['neighbor_dists'], expected_dists)

    def test_uid_nearest_neighbors_no_uid(self):
        """ Test not providing base64 """
        data = dict(
            # uid='some-uid',
            k=10,
        )
        r = self.app.test_client().post('/uid_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, 'No UID provided')

    def test_uid_nearest_neighbors_no_k(self):
        """ Test not providing base64 """
        data = dict(
            uid='some-uid',
            # k=10,
        )
        r = self.app.test_client().post('/uid_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "No 'k' value provided")

    def test_uid_nearest_neighbors_bad_k_json(self):
        """ Test string for k """
        data = dict(
            uid='some-uid',
            k="10.2",  # float string fails an int cast.
        )
        r = self.app.test_client().post('/uid_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Failed to convert 'k' argument to an "
                                       "integer")

    def test_uid_nearest_neighbors_no_elem_for_uid(self):
        def raise_keyerror(*_):
            raise KeyError("invalid-key")
        self.app.descriptor_index.get_descriptor = mock.Mock(
            side_effect=raise_keyerror
        )

        data = dict(
            uid='some-uid',
            k=3,
        )
        r = self.app.test_client().post('/uid_nearest_neighbors',
                                        data=data)
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Failed to get descriptor for UID "
                                       "some-uid")

    def test_uid_nearest_neighbors(self):
        expected_uids = ['a', 'b', 'c']
        expected_neighbors = []
        for uid in expected_uids:
            e = mock.Mock(spec=DescriptorElement)
            e.uuid.return_value = uid
            expected_neighbors.append(e)
        expected_dists = [1, 2, 3]

        self.app.descriptor_index.get_descriptor = mock.Mock()
        self.app.neighbor_index.nn = mock.Mock(
            return_value=[expected_neighbors, expected_dists]
        )

        data = dict(
            uid='some-uid',
            k=10,
        )
        r = self.app.test_client().post('/uid_nearest_neighbors',
                                        data=data)
        self.app.descriptor_index.get_descriptor.assert_called_once_with(
            data['uid']
        )
        self.app.neighbor_index.nn.assert_called_once_with(
            self.app.descriptor_index.get_descriptor(), data['k']
        )

        self.assertStatusCode(r, 200)
        r_json = json.loads(r.data.decode())
        self.assertListEqual(r_json['neighbor_uids'], expected_uids)
        self.assertListEqual(r_json['neighbor_dists'], expected_dists)

    def test_refine_no_session_id(self):
        with self.app.test_client() as tc:
            r = tc.post('/refine')
            self.assertStatusCode(r, 400)
            self.assertJsonMessageRegex(r, "No session id \(sid\) provided")

    def test_refine_invalid_session_id(self):
        with self.app.test_client() as tc:
            r = tc.post('/refine',
                        data={
                            'sid': 'invalid-sid'
                        })
            self.assertStatusCode(r, 404)
            self.assertJsonMessageRegex(r, "session id invalid-sid not found")

    def test_get_iqr_state_no_sid(self):
        # Test that calling GET /state with no SID results in error.
        r = self.app.test_client().get('/state')
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, 'No session id')

    def test_get_iqr_state_bad_sid_empty_controller(self):
        # Test that giving an invalid SID to get_iqr_state results in error
        # message return, when no sessions in controller.

        # App should currently have no sessions.
        self.assertTupleEqual(
            self.app.controller.session_uuids(),
            ()
        )

        # Attempt getting a state with an SID not currently registered in the
        # controller.
        r = self.app.test_client().get('/state',
                                       query_string=dict(
                                           sid='something-invalid'
                                       ))
        self.assertStatusCode(r, 404)
        self.assertJsonMessageRegex(r, 'not found')

    def test_get_iqr_state_base_sid_nonempty_controller(self):
        # Test that giving an invalid SID to get_iqr_state results in error
        # message return when controller has one or more sessions in it.

        # Initialize two different sessions
        self.app.test_client().post('/session',
                                    data=dict(
                                        sid='test-session-1'
                                    ))
        self.app.test_client().post('/session',
                                    data=dict(
                                        sid='test-session-2'
                                    ))

        # attempt to get state for a session not created
        r = self.app.test_client().get('/state',
                                       query_string=dict(
                                           sid='test-session-3'
                                       ))
        self.assertStatusCode(r, 404)
        self.assertJsonMessageRegex(r, 'not found')

    def test_get_iqr_state(self):
        # Test that the base64 returned is what is returned from
        #  IqrSession.get_state_bytes (mocked)
        expected_bytes = b'these-bytes'
        expected_b64 = base64.b64encode(expected_bytes).decode()

        # Pretend input SID is valid
        self.app.controller.has_session_uuid = mock.Mock(return_value=True)
        self.app.controller.get_session = mock.Mock()
        self.app.controller.get_session().get_state_bytes.return_value = \
            expected_bytes

        r = self.app.test_client().get('/state',
                                       query_string=dict(
                                           sid='some-sid'
                                       ))
        self.assertStatusCode(r, 200)
        r_json = json.loads(r.data.decode())
        self.assertEqual(r_json['message'], "Success")
        self.assertEqual(r_json['sid'], 'some-sid')
        self.assertEqual(r_json['state_b64'], expected_b64)

    def test_set_iqr_state_no_sid(self):
        # Test that calling set_iqr_state with no SID returns an error
        r = self.app.test_client().put('/state',
                                       data=dict(
                                           state_base64='dummy'
                                       ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "No session id \(sid\) provided")

    def test_set_iqr_state_no_b64(self):
        """
        Test that calling set_iqr_state with no base_64 data returns an
        error.
        """
        r = self.app.test_client().put('/state',
                                       data=dict(
                                           sid='dummy'
                                       ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, 'No state package base64 provided')

    def test_set_iqr_state_invalid_base64(self):
        # Test when PUT /state is given invalid base64 data.

        r = self.app.test_client().put('/state',
                                       data=dict(
                                           sid='test-sid',
                                           state_base64='some-invalid-data'
                                       ))
        self.assertStatusCode(r, 400)
        self.assertJsonMessageRegex(r, "Invalid base64 input")

    def test_set_iqr_state_invalid_sid(self):
        # Test that an invalid SID provided causes an error.  Must set a state
        # to an existing session.
        test_b64 = base64.b64encode(b'test')

        # App starts with nothing in session controller, so no SID should be
        # initially valid.
        r = self.app.test_client().put('/state',
                                       data=dict(
                                           sid='invalid',
                                           state_base64=test_b64
                                       ))
        self.assertStatusCode(r, 404)
        self.assertJsonMessageRegex(r, 'not found')

    def test_set_iqr_state(self):
        # Test that expected base64 decoded bytes are passed to
        # `IqrSession.set_state_bytes` method.
        expected_sid = 'test-sid'
        expected_bytes = b'expected-state-bytes'
        expected_bytes_b64 = base64.b64encode(expected_bytes).decode()

        self.app.controller.has_session_uuid = mock.Mock(return_value=True)
        self.app.controller.get_session = mock.Mock()

        r = self.app.test_client().put('/state',
                                       data=dict(
                                           sid=expected_sid,
                                           state_base64=expected_bytes_b64
                                       ))

        self.app.controller.get_session().set_state_bytes.assert_called_with(
            expected_bytes, self.app.descriptor_factory
        )
        self.assertStatusCode(r, 200)
        r_json = json.loads(r.data.decode())
        self.assertRegexpMatches(r_json['message'], 'Success')
        self.assertRegexpMatches(r_json['sid'], expected_sid)
